package ru.dbotthepony.mc.otm.datagen.models

import com.google.gson.JsonObject
import com.google.gson.JsonParser
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.packs.PackType
import net.minecraft.world.level.block.Block
import net.minecraftforge.client.model.generators.ModelProvider
import net.minecraftforge.forge.event.lifecycle.GatherDataEvent
import ru.dbotthepony.mc.otm.datagen.DataGen

private typealias Callback = (MatteryModelProvider) -> Unit

sealed class MatteryModelProvider(event: GatherDataEvent, folder: String) : ModelProvider<MatteryModelBuilder>(event.generator, DataGen.MOD_ID, folder, ::MatteryModelBuilder, event.existingFileHelper) {
	private fun extendWithFolder(rl: ResourceLocation): ResourceLocation {
		return if (rl.path.contains("/")) rl else ResourceLocation(rl.namespace, folder + "/" + rl.path)
	}

	fun copy(destination: String, source: String): MatteryModelBuilder {
		val destinationLocation = extendWithFolder(if (destination.contains(":")) ResourceLocation(destination) else ResourceLocation(modid, destination))
		val sourceLocation = extendWithFolder(if (source.contains(":")) ResourceLocation("models/$source.json") else ResourceLocation(modid, "models/$source.json"))

		check(!generatedModels.containsKey(destinationLocation)) { "Model provider already contains model $destinationLocation" }
		existingFileHelper.trackGenerated(destinationLocation, MODEL)

		return factory.apply(destinationLocation).also {
			generatedModels[destinationLocation] = it
			val resource = existingFileHelper.getResource(sourceLocation, PackType.CLIENT_RESOURCES)
			val stream = resource.inputStream
			val reader = stream.reader()

			try {
				it.fromJson(JsonParser.parseReader(reader) as JsonObject)
			} finally {
				reader.close()
				stream.close()
			}
		}
	}

	private val callbacks = ArrayList<Callback>()

	fun lambda(callback: Callback) {
		callbacks.add(callback)
	}

	override fun registerModels() {
		for (callback in callbacks) {
			callback(this)
		}
	}
}

private data class CubeAllEntry(val path: String, val texture: ResourceLocation)
private data class CubeColumnEntry(val path: String, val end: ResourceLocation, val side: ResourceLocation)

class BlockMatteryModelProvider(event: GatherDataEvent) : MatteryModelProvider(event, BLOCK_FOLDER) {
	override fun getName(): String {
		return "Block Models: $modid"
	}

	private val cubeAll = ArrayList<CubeAllEntry>()
	private val cubeColumn = ArrayList<CubeColumnEntry>()

	override fun registerModels() {
		super.registerModels()

		for ((path, texture) in cubeAll) {
			cubeAll(path, texture)
		}

		for ((path, end, side) in cubeColumn) {
			cubeColumn(path, end, side)
		}
	}

	fun decorativeCubeAll(vararg blocks: Block) {
		for (it in blocks) {
			cubeAll.add(CubeAllEntry(it.registryName!!.path, ResourceLocation("overdrive_that_matters:block/decorative/${it.registryName!!.path}")))
		}
	}

	fun decorativeCubeAll(blocks: List<Block>) {
		for (it in blocks) {
			cubeAll.add(CubeAllEntry(it.registryName!!.path, ResourceLocation("overdrive_that_matters:block/decorative/${it.registryName!!.path}")))
		}
	}

	fun column(it: Block, end: String, side: String) {
		cubeColumn.add(CubeColumnEntry(it.registryName!!.path, ResourceLocation(DataGen.MOD_ID, end), ResourceLocation(DataGen.MOD_ID, side)))
	}

	fun decorativeColumn(it: Block, end: String, side: String) {
		column(it, "block/decorative/$end", "block/decorative/$side")
	}

	fun resourceCubeAll(vararg blocks: Block) {
		for (it in blocks) {
			cubeAll.add(CubeAllEntry(it.registryName!!.path, ResourceLocation("overdrive_that_matters:block/resource/${it.registryName!!.path}")))
		}
	}
}
