package ru.dbotthepony.mc.otm.client.render;

import com.mojang.blaze3d.systems.RenderSystem;
import com.mojang.blaze3d.vertex.*;
import com.mojang.math.Matrix4f;
import net.minecraft.client.Minecraft;
import net.minecraft.client.renderer.GameRenderer;
import net.minecraft.resources.ResourceLocation;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.core.RGBAColor;

import java.util.Stack;

import static org.lwjgl.opengl.GL11.GL_ALWAYS;

/**
 * I am too lazy to learn how Mojang's API works
 * so I just recreate part of GMod's API in here
 *
 * also contains texture widgets
 */
public class RenderHelper {
	private static final Matrix4f identity = new Matrix4f();
	public static final ResourceLocation WIDGETS = new ResourceLocation(OverdriveThatMatters.MOD_ID, "textures/gui/widgets.png");

	static {
		identity.setIdentity();
	}

	public static float depth_z = 0;

	public static void setUpcomingDepth(float depth) {
		depth_z = depth;
	}

	public static void clearUpcomingDepth() {
		depth_z = 0;
	}

	// Regular functions

	/**
	 * Draws a textured rectangle on screen
	 * translated by a given matrix
	 *
	 * @param matrix Matrix translation to use
	 * @param x screen X position (with matrix transformation)
	 * @param y screen Y position (with matrix transformation)
	 * @param width width of drawn rectangle
	 * @param height height of drawn rectangle
	 * @param u0 initial U position ranging from 0-1 (single) or beyond (repetitive)
	 * @param v0 initial V position ranging from 0-1 (single) or beyond (repetitive)
	 * @param u1 final U position ranging from 0-1 (single) or beyond (repetitive)
	 * @param v1 final V position ranging from 0-1 (single) or beyond (repetitive)
	 */
	public static void drawTexturedRectUV(
		Matrix4f matrix,
		float x,
		float y,
		float width,
		float height,
		float u0,
		float v0,
		float u1,
		float v1
	) {
		RenderSystem.setShader(GameRenderer::getPositionTexShader);
		RenderSystem.enableTexture();
		RenderSystem.enableBlend();
		RenderSystem.defaultBlendFunc();

		RenderSystem.depthFunc(GL_ALWAYS);

		var builder = Tesselator.getInstance().getBuilder();
		builder.begin(VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION_TEX);
		builder.vertex(matrix, x, y + height, depth_z).uv(u0, v1).endVertex();
		builder.vertex(matrix, x + width, y + height, depth_z).uv(u1, v1).endVertex();
		builder.vertex(matrix, x + width, y, depth_z).uv(u1, v0).endVertex();
		builder.vertex(matrix, x, y, depth_z).uv(u0, v0).endVertex();
		builder.end();
		BufferUploader.end(builder);
	}

	public static void drawTexturedRectUV(
		Matrix4f matrix,
		float x,
		float y,
		float width,
		float height,
		UVCoords uv
	) {
		drawTexturedRectUV(matrix, x, y, width, height, uv.u0(), uv.v0(), uv.u1(), uv.v1());
	}

	public static void colorSphere(Matrix4f matrix, float radius) {
		final int fragments = 32;

		RenderSystem.enableBlend();
		RenderSystem.defaultBlendFunc();

		var builder = Tesselator.getInstance().getBuilder();
		builder.begin(VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION_COLOR);

		final double turn_step = Math.PI * (1 / (double) fragments) * 2;
		final double stripe_step = Math.PI * (1 / (double) fragments);

		for (int turn = 0; turn < fragments; turn++) {
			final double turn_pre = turn_step * turn;
			final double turn_post = turn_step * (turn + 1);

			for (int sector = 0; sector < fragments; sector++) {
				final double tilt_pre = Math.PI / 2 - stripe_step * (sector);
				final double tilt_post = Math.PI / 2 - stripe_step * (sector + 1);

				final float x_pre = (float) Math.sin(turn_pre) * radius;
				final float x_post = (float) Math.sin(turn_post) * radius;

				final float z_pre = (float) Math.cos(turn_pre) * radius;
				final float z_post = (float) Math.cos(turn_post) * radius;

				final float y_pre = (float) (Math.sin(tilt_pre) + 0.5) * radius;
				final float y_post = (float) (Math.sin(tilt_post) + 0.5) * radius;

				builder.vertex(matrix, x_pre * (float) Math.cos(tilt_post), y_post, z_pre * (float) Math.cos(tilt_post))
					.color(draw_color.getRed(), draw_color.getGreen(), draw_color.getBlue(), draw_color.getAlpha())
					.endVertex();

				builder.vertex(matrix, x_post * (float) Math.cos(tilt_post), y_post, z_post * (float) Math.cos(tilt_post))
					.color(draw_color.getRed(), draw_color.getGreen(), draw_color.getBlue(), draw_color.getAlpha())
					.endVertex();

				builder.vertex(matrix, x_post * (float) Math.cos(tilt_pre), y_pre, z_post * (float) Math.cos(tilt_pre))
					.color(draw_color.getRed(), draw_color.getGreen(), draw_color.getBlue(), draw_color.getAlpha())
					.endVertex();

				builder.vertex(matrix, x_pre * (float) Math.cos(tilt_pre), y_pre, z_pre * (float) Math.cos(tilt_pre))
					.color(draw_color.getRed(), draw_color.getGreen(), draw_color.getBlue(), draw_color.getAlpha())
					.endVertex();

			}
		}

		builder.end();
		BufferUploader.end(builder);
	}

	public static void colorSphere(PoseStack stack, float radius) {
		colorSphere(stack.last().pose(), radius);
	}

	/**
	 * Draws a textured rectangle on screen
	 *
	 * @param x screen X position (with matrix transformation)
	 * @param y screen Y position (with matrix transformation)
	 * @param width width of drawn rectangle
	 * @param height height of drawn rectangle
	 * @param u0 initial U position ranging from 0-1 (single) or beyond (repetitive)
	 * @param v0 initial V position ranging from 0-1 (single) or beyond (repetitive)
	 * @param u1 final U position ranging from 0-1 (single) or beyond (repetitive)
	 * @param v1 final V position ranging from 0-1 (single) or beyond (repetitive)
	 */
	public static void drawTexturedRectUV(
		float x,
		float y,
		float width,
		float height,
		float u0,
		float v0,
		float u1,
		float v1
	) {
		drawTexturedRectUV(identity, x, y, width, height, u0, v0, u1, v1);
	}

	public static void drawTexturedRectUV(
		float x,
		float y,
		float width,
		float height,
		UVCoords uv
	) {
		drawTexturedRectUV(identity, x, y, width, height, uv.u0(), uv.v0(), uv.u1(), uv.v1());
	}

	/**
	 * Draws a textured rectangle on screen
	 * with matrix from pose
	 *
	 * @param stack PoseStack containing translation matrix
	 * @param x screen X position (with matrix transformation)
	 * @param y screen Y position (with matrix transformation)
	 * @param width width of drawn rectangle
	 * @param height height of drawn rectangle
	 * @param u0 initial U position ranging from 0-1 (single) or beyond (repetitive)
	 * @param v0 initial V position ranging from 0-1 (single) or beyond (repetitive)
	 * @param u1 final U position ranging from 0-1 (single) or beyond (repetitive)
	 * @param v1 final V position ranging from 0-1 (single) or beyond (repetitive)
	 */
	public static void drawTexturedRectUV(
		PoseStack stack,
		float x,
		float y,
		float width,
		float height,
		float u0,
		float v0,
		float u1,
		float v1
	) {
		drawTexturedRectUV(stack.last().pose(), x, y, width, height, u0, v0, u1, v1);
	}

	public static void drawTexturedRectUV(
		PoseStack stack,
		float x,
		float y,
		float width,
		float height,
		UVCoords uv
	) {
		drawTexturedRectUV(stack.last().pose(), x, y, width, height, uv.u0(), uv.v0(), uv.u1(), uv.v1());
	}

	/**
	 * @param stack stack to take transformation matrix from
	 * @param x position x on screen
	 * @param y position y on screen
	 * @param width drawn rectangle width
	 * @param height drawn rectangle height
	 * @param image_x absolute asset x, to compute U0 coordinate from
	 * @param image_y absolute asset y, to compute V0 coordinate from
	 * @param mapped_width absolute asset width, to compute U1 coordinate from
	 * @param mapped_height absolute asset height, to compute V1 coordinate from
	 */
	public static void drawTexturedRectAuto(
		PoseStack stack,
		float x,
		float y,
		float width,
		float height,
		float image_x,
		float image_y,
		float mapped_width,
		float mapped_height
	) {
		drawTexturedRectUV(stack.last().pose(), x, y, width, height, image_x / mapped_width, image_y / mapped_height, (image_x + width) / mapped_width, (image_y + height) / mapped_height);
	}

	public static void drawTexturedRectAuto(
		PoseStack stack,
		float x,
		float y,
		float width,
		float height,
		float image_x,
		float image_y,
		float mapped_width,
		float mapped_height,
		UVWindingOrder order
	) {
		drawTexturedRectUV(stack.last().pose(), x, y, width, height, order.translate(image_x / mapped_width, image_y / mapped_height, (image_x + width) / mapped_width, (image_y + height) / mapped_height));
	}

	public static void drawTexturedRectAuto(
		PoseStack stack,
		float x,
		float y,
		float width,
		float height,
		float image_x,
		float image_y
	) {
		drawTexturedRectAuto(stack, x, y, width, height, image_x, image_y, 256, 256);
	}

	public static void drawTexturedRectAuto(
		PoseStack stack,
		float x,
		float y,
		float width,
		float height,
		float image_x,
		float image_y,
		UVWindingOrder order
	) {
		drawTexturedRectAuto(stack, x, y, width, height, image_x, image_y, 256, 256, order);
	}

	/**
	 * Draws a textured rectangle on screen
	 * with matrix from pose
	 * UVs are 0-1
	 *
	 * @param stack PoseStack containing translation matrix
	 * @param x screen X position (with matrix transformation)
	 * @param y screen Y position (with matrix transformation)
	 * @param width width of drawn rectangle
	 * @param height height of drawn rectangle
	 */
	public static void drawTexturedRect(
		PoseStack stack,
		float x,
		float y,
		float width,
		float height
	) {
		drawTexturedRectUV(stack, x, y, width, height, 0, 0, 1, 1);
	}

	public static void drawRegularSlot(
		PoseStack stack,
		float x,
		float y
	) {
		RenderSystem.setShaderTexture(0, WIDGETS);
		drawTexturedRectAuto(stack, x, y, 18, 18, 0, 96, 256, 256);
	}

	public static void drawBigSlot(
		PoseStack stack,
		float x,
		float y
	) {
		RenderSystem.setShaderTexture(0, WIDGETS);
		drawTexturedRectAuto(stack, x, y, 26, 26, 18, 96, 256, 256);
	}

	public static RGBAColor getDrawColor() {
		return draw_color;
	}

	public static void setDrawColor(RGBAColor rect_color) {
		RenderHelper.draw_color = rect_color;
	}

	public static void setShaderColor(RGBAColor rect_color) {
		RenderSystem.setShaderColor(rect_color.getRed(), rect_color.getGreen(), rect_color.getBlue(), rect_color.getAlpha());
	}

	private static RGBAColor draw_color = new RGBAColor(255, 255, 255, 255);

	/**
	 * Draws a solid color rectangle on screen, defined by last setDrawColor(RGBColor rect_color) call
	 *
	 * @param matrix matrix to translate by
	 * @param x x screen coordinate
	 * @param y y screen coordinate
	 * @param width rectangle width
	 * @param height rectangle height
	 */
	public static void drawRect(
		Matrix4f matrix,
		float x,
		float y,
		float width,
		float height
	) {
		RenderSystem.disableTexture();
		RenderSystem.enableBlend();
		RenderSystem.defaultBlendFunc();
		RenderSystem.setShader(GameRenderer::getPositionColorShader);

		RenderSystem.depthFunc(GL_ALWAYS);

		var tess = Tesselator.getInstance();
		BufferBuilder builder = tess.getBuilder();

		builder.begin(VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION_COLOR);

		builder.vertex(matrix, x, y + height, depth_z).color(draw_color.getRed(), draw_color.getGreen(), draw_color.getBlue(), draw_color.getAlpha()).endVertex();
		builder.vertex(matrix, x + width, y + height, depth_z).color(draw_color.getRed(), draw_color.getGreen(), draw_color.getBlue(), draw_color.getAlpha()).endVertex();
		builder.vertex(matrix, x + width, y, depth_z).color(draw_color.getRed(), draw_color.getGreen(), draw_color.getBlue(), draw_color.getAlpha()).endVertex();
		builder.vertex(matrix, x, y, depth_z).color(draw_color.getRed(), draw_color.getGreen(), draw_color.getBlue(), draw_color.getAlpha()).endVertex();

		tess.end();

		RenderSystem.enableTexture();
	}

	/**
	 * Draws a solid color rectangle on screen, defined by last setDrawColor(RGBColor rect_color) call
	 *
	 * @param stack stack to translate by
	 * @param x x screen coordinate
	 * @param y y screen coordinate
	 * @param width rectangle width
	 * @param height rectangle height
	 */
	public static void drawRect(
		PoseStack stack,
		float x,
		float y,
		float width,
		float height
	) {
		drawRect(stack.last().pose(), x, y, width, height);
	}

	record ScissorRect(int x, int y, int width, int height) {}

	private static final Stack<ScissorRect> SCISSOR = new Stack<>();

	public static void ensureScissorStackEmpty() {
		if (!SCISSOR.empty()) {
			throw new IllegalStateException("Unbalanced scissor rects: Popping less than pushing");
		}
	}

	public static void pushScissorRect(int x, int y, int width, int height) {
		if (!SCISSOR.empty()) {
			final var peek = SCISSOR.peek();
			x = Math.max(x, peek.x);
			y = Math.max(y, peek.y);
			width = Math.min(width, peek.width);
			height = Math.min(height, peek.height);
		}

		SCISSOR.push(new ScissorRect(x, y, width, height));

		final var window = Minecraft.getInstance().getWindow();
		y = window.getHeight() - y - height;

		RenderSystem.enableScissor(x, y, width, height);
	}

	public static void popScissorRect() {
		if (SCISSOR.empty()) {
			throw new IllegalStateException("Unbalanced scissor rects: Popping more than pushing");
		}

		SCISSOR.pop();

		if (SCISSOR.empty()) {
			RenderSystem.disableScissor();
			return;
		}

		final var value = SCISSOR.peek();
		final var window = Minecraft.getInstance().getWindow();
		final var y = window.getHeight() - value.y - value.height;
		RenderSystem.enableScissor(value.x, y, value.width, value.height);
	}
}
