package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.CapabilityItemHandler
import ru.dbotthepony.mc.otm.block.MatterBottlerBlock
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerState
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler
import ru.dbotthepony.mc.otm.capability.matter.MatterDirection
import ru.dbotthepony.mc.otm.capability.matter.MatterHandlerImpl
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.menu.MatterBottlerMenu
import ru.dbotthepony.mc.otm.orNull
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.set

class MatterBottlerBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) :
	MatteryPoweredBlockEntity(MBlockEntities.MATTER_BOTTLER, p_155229_, p_155230_), IMatterGraphNode {

	private val node = Graph6Node<IMatterGraphNode>(this)
	private val resolverNode = LazyOptional.of { this }
	override val energy = WorkerEnergyStorage(this)

	override fun getAsMatterNode(): Graph6Node<IMatterGraphNode> {
		return node
	}

	// true - bottling
	// false - unbottling
	var workFlow: Boolean = true
		set(value) {
			field = value
			this.setChangedLight()
		}

	fun switchWorkFlow() {
		workFlow = !workFlow
		updateBlockState()
	}

	private fun updateBlockState() {
		val level = level as? ServerLevel ?: return

		var state = blockState
		val initial = if (workFlow) 0 else 3

		for (i in initial .. initial + 2) {
			val desired = !container.getItem(i).isEmpty && container.getItem(i).getCapability(MatteryCapability.MATTER).isPresent

			if (state.getValue(MatterBottlerBlock.SLOT_PROPERTIES[i - initial]) != desired) {
				state = state.setValue(MatterBottlerBlock.SLOT_PROPERTIES[i - initial], desired)
			}
		}

		if (state !== blockState) {
			level.setBlock(blockPos, state, Block.UPDATE_CLIENTS)
		}
	}

	// TODO: оно должно что то делать
	// true - continue even if empty when bottling / if full while unbottling
	// false - spit into output slot
	//private var work_behavior = true

	@JvmField
	val container: MatteryContainer = object : MatteryContainer(this::setChangedLight, 6) {
		override fun getMaxStackSize(slot: Int): Int {
			return 1
		}

		override fun setChanged(slot: Int, new: ItemStack, old: ItemStack) {
			super.setChanged(slot, new, old)
			updateBlockState()
		}
	}

	val itemHandler = container.handler({ slot: Int, stack: ItemStack ->
		if (this.workFlow) {
			return@handler slot < 3 && stack.getCapability(MatteryCapability.MATTER).isPresent
		}
		slot >= 3 && stack.getCapability(MatteryCapability.MATTER).isPresent
	}, { slot: Int, amount: Int, stack: ItemStack? ->
		if (this.workFlow) {
			return@handler slot >= 3
		}
		slot < 3
	})

	@JvmField
	val matter: MatterHandlerImpl =
		object : MatterHandlerImpl(this::setChangedLight, MatterDirection.BIDIRECTIONAL, CAPACITY) {
			override val direction: MatterDirection get() {
				return if (this@MatterBottlerBlockEntity.workFlow) MatterDirection.RECEIVE else MatterDirection.EXTRACT
			}
		}

	private var initialCapacity: ImpreciseFraction? = null
	private var lastWorkStack: ItemStack? = null

	override fun getMatterHandler(): IMatterHandler {
		return matter
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			MatterNetworkGraph.discoverFull(this, node)
	}

	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		itemHandler.invalidate()
		matter.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		itemHandler.revive()
		matter.revive()
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid) {
			if (cap === MatteryCapability.MATTER) {
				return matter.get().cast()
			}

			if (cap === CapabilityItemHandler.ITEM_HANDLER_CAPABILITY) {
				return itemHandler.get().cast()
			}

			if (cap === MatteryCapability.MATTER_NODE) {
				return resolverNode.cast()
			}
		}

		return super.getCapability(cap, side)
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterBottlerMenu(containerID, inventory, this)
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["work_slots"] = container.serializeNBT()
		nbt["work_flow"] = this.workFlow
		nbt["matter"] = matter.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		container.deserializeNBT(nbt["work_slots"])
		workFlow = nbt.getBoolean("work_flow")
		nbt.ifHas("matter", CompoundTag::class.java, matter::deserializeNBT)
	}

	fun getWorkProgress(): Float {
		val lastWorkStack = lastWorkStack ?: return 0f
		val initialCapacity = initialCapacity ?: return 0f
		val cap = lastWorkStack.getCapability(MatteryCapability.MATTER).orNull() ?: return 0f

		if (this.workFlow) {
			if (cap.maxStoredMatter - initialCapacity <= ImpreciseFraction.ZERO) {
				return 0f
			}

			return ((cap.storedMatter - initialCapacity) / (cap.maxStoredMatter - initialCapacity)).toFloat()
		}

		if (initialCapacity <= ImpreciseFraction.ZERO) {
			return 0f
		}

		return (ImpreciseFraction.ONE - cap.storedMatter / initialCapacity).toFloat()
	}

	override fun setRemoved() {
		super.setRemoved()
		node.destroy(::MatterNetworkGraph)
	}

	fun tick() {
		batteryChargeLoop()

		if (isBlockedByRedstone) {
			if (blockState.getValue(WorkerState.SEMI_WORKER_STATE) !== WorkerState.IDLE) {
				level?.setBlock(
					blockPos,
					blockState.setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.IDLE),
					Block.UPDATE_CLIENTS
				)
			}

			return
		}

		var work_stack: ItemStack? = null
		var capability: IMatterHandler? = null
		val align = if (workFlow) 0 else 3
		var work_slot = -1
		val unexpectedDirection = if (workFlow) MatterDirection.EXTRACT else MatterDirection.RECEIVE

		for (i in align until align + 3) {
			val itemStack = container.getItem(i)

			if (!itemStack.isEmpty) {
				val cap = itemStack.getCapability(MatteryCapability.MATTER).orNull() ?: continue

				if (cap.direction !== unexpectedDirection) {
					if (this.workFlow && cap.missingMatter > ImpreciseFraction.ZERO || !this.workFlow && cap.storedMatter > ImpreciseFraction.ZERO) {
						work_stack = itemStack
						capability = cap
						work_slot = i
						break
					}
				}
			}
		}

		if (work_stack == null) {
			lastWorkStack = null
			initialCapacity = null
		} else if (work_stack != lastWorkStack) {
			lastWorkStack = work_stack
			initialCapacity = capability!!.storedMatter
		}

		val graph = node.graph as MatterNetworkGraph?

		if (capability != null) {
			if (blockState.getValue(WorkerState.SEMI_WORKER_STATE) !== WorkerState.WORKING) {
				level?.setBlock(blockPos, blockState.setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS)
			}

			if (workFlow) {
				if (matter.storedMatter < MATTER_EXCHANGE_RATE && graph != null) {
					val extracted = graph.extractMatter(
						matter.missingMatter.coerceAtMost(MATTER_EXCHANGE_RATE * EXTRACTION_TICKS).coerceAtMost(capability.missingMatter - matter.storedMatter), true
					)

					if (extracted > ImpreciseFraction.ZERO) {
						val received = matter.receiveMatterOuter(extracted, false)
						graph.extractMatter(received, false)
					}
				}

				if (matter.storedMatter > ImpreciseFraction.ZERO) {
					val energyExtracted = energy.extractEnergyInner(ENERGY_CONSUMPTION, true)

					if (!energyExtracted.isZero) {
						val matter = capability.receiveMatterOuter(MATTER_EXCHANGE_RATE.coerceAtMost(matter.storedMatter) * energyExtracted / ENERGY_CONSUMPTION, true)

						if (!matter.isZero) {
							energy.extractEnergyInner(ENERGY_CONSUMPTION * matter / MATTER_EXCHANGE_RATE, false)

							capability.receiveMatterOuter(matter, false)
							this.matter.extractMatterInner(matter, false)

							if (capability.missingMatter.isZero) {
								for (i in 3..5) {
									if (container.getItem(i).isEmpty) {
										container.setItem(work_slot, ItemStack.EMPTY)
										container.setItem(i, work_stack!!)
										break
									}
								}
							}
						}
					}
				}
			} else {
				val energyExtracted = energy.extractEnergyInner(ENERGY_CONSUMPTION, true)

				if (!energyExtracted.isZero) {
					val matter = capability.extractMatterOuter(MATTER_EXCHANGE_RATE.coerceAtMost(matter.missingMatter) * energyExtracted / ENERGY_CONSUMPTION, true)

					if (!matter.isZero) {
						this.energy.extractEnergyInner(ENERGY_CONSUMPTION * matter / MATTER_EXCHANGE_RATE,false)

						capability.extractMatterOuter(matter, false)
						this.matter.receiveMatterInner(matter, false)

						if (capability.storedMatter.isZero) {
							for (i in 2 downTo 0) {
								if (container.getItem(i).isEmpty) {
									container.setItem(work_slot, ItemStack.EMPTY)
									container.setItem(i, work_stack!!)
									break
								}
							}
						}
					}
				}
			}
		} else {
			level!!.setBlock(blockPos, blockState.setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
		}

		if (!workFlow && !matter.storedMatter.isZero && graph != null) {
			val diff = matter.extractMatterInner(matter.storedMatter, true)
			val diff2 = graph.receiveMatter(diff, true)
			matter.extractMatterInner(diff2, false)
			graph.receiveMatter(diff2, false)
		}
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.matter_bottler")
		private val MATTER_EXCHANGE_RATE = ImpreciseFraction("0.04")
		private val ENERGY_CONSUMPTION = ImpreciseFraction(20)
		private val EXTRACTION_TICKS = ImpreciseFraction(200)
		private val CAPACITY = ImpreciseFraction(4)
	}
}