package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.capability.matter.IMatterTaskProvider
import ru.dbotthepony.mc.otm.menu.MatterPanelMenu
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import java.util.HashMap
import java.util.UUID
import ru.dbotthepony.mc.otm.capability.matter.MatterTask
import java.util.stream.Collectors
import ru.dbotthepony.mc.otm.capability.matter.MatterTaskAllocation
import ru.dbotthepony.mc.otm.capability.matter.PatternState
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.level.Level
import net.minecraftforge.common.capabilities.Capability
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import java.util.ArrayList
import java.util.List

class MatterPanelBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) :
	MatteryBlockEntity(MBlockEntities.MATTER_PANEL, p_155229_, p_155230_), IMatterGraphNode, IMatterTaskProvider {

	private val listeners = ArrayList<MatterPanelMenu>()
	private val node = Graph6Node<IMatterGraphNode>(this)

	override fun getAsMatterNode(): Graph6Node<IMatterGraphNode> {
		return node
	}

	fun attachMenu(menu: MatterPanelMenu) {
		listeners.add(menu)
	}

	fun deatachMenu(menu: MatterPanelMenu) {
		listeners.remove(menu)
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterPanelMenu(containerID, inventory, this)
	}

	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		resolver.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		resolver = LazyOptional.of { this }
	}

	private var resolver = LazyOptional.of { this }

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid && (cap === MatteryCapability.MATTER_NODE || cap === MatteryCapability.TASK))
			return resolver.cast()

		return super.getCapability(cap, side)
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			MatterNetworkGraph.discoverFull(this, node)
	}

	override fun setRemoved() {
		super.setRemoved()
		node.destroy(::MatterNetworkGraph)
	}

	override fun getTaskHandler(): IMatterTaskProvider? {
		return this
	}

	private val _tasks = HashMap<UUID, MatterTask>()

	override val tasks: Collection<MatterTask> get() {
		return _tasks.values.stream().filter { task: MatterTask? -> task!!.required() > 0 }.collect(Collectors.toList()) as Collection<MatterTask>
	}

	override val allTasks: Collection<MatterTask> get() {
		return List.copyOf(_tasks.values) as Collection<MatterTask>
	}

	override fun allocateTask(simulate: Boolean): MatterTaskAllocation? {
		val graph = node.graph as MatterNetworkGraph? ?: return null

		for ((key, task) in _tasks) {
			if (task.required > 0) {
				val getPattern = graph.getPattern(task.pattern!!)

				if (getPattern != null) {
					if (!simulate) {
						val new = task.shrinkRequired(1)
						_tasks[key] = new
						listeners.forEach { menu: MatterPanelMenu -> menu.taskUpdated(new) }
						graph.onMatterTaskUpdated(new, task)
						setChanged()
					}

					return MatterTaskAllocation(task, getPattern)
				}
			}
		}

		return null
	}

	override fun notifyTaskCompletion(task: MatterTask): Boolean {
		var localTask = _tasks[task.id] ?: return false
		val oldTask = localTask

		localTask = localTask.shrinkInProgress(1)
		val graph = node.graph as MatterNetworkGraph?

		// Задача полностью выполнена
		if (localTask.required <= 0 && localTask.in_progress <= 0) {
			_tasks.remove(task.id)
			graph?.onMatterTaskCreated(task)
			listeners.forEach { menu: MatterPanelMenu -> menu.taskRemoved(localTask) }
		} else {
			// Задача обновлена
			_tasks[task.id()] = localTask
			graph?.onMatterTaskUpdated(localTask, oldTask)
			listeners.forEach { menu: MatterPanelMenu -> menu.taskUpdated(localTask) }
		}

		setChanged()
		return true
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		val list = ListTag()

		for (task in _tasks.values) {
			list.add(task!!.serializeNBT())
		}

		nbt.put("tasks", list)
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		_tasks.clear()
		val list = nbt.getList("tasks", Tag.TAG_COMPOUND.toInt())

		for (tag in list) {
			val task = MatterTask.deserializeNBT(tag)

			if (task != null) {
				_tasks[task.id()] = task
			}
		}
	}

	override fun getTask(id: UUID): MatterTask? {
		return _tasks[id]
	}

	fun removeTask(id: UUID) {
		val task = _tasks[id] ?: return
		_tasks.remove(id)

		(node.graph as MatterNetworkGraph?)?.onMatterTaskRemoved(task)

		listeners.forEach { menu: MatterPanelMenu -> menu.taskRemoved(task) }
		setChanged()
	}

	fun removeTask(state: PatternState) = removeTask(state.id)

	fun addTask(state: PatternState, how_much: Int): MatterTask {
		val task = MatterTask(UUID.randomUUID(), state.id, state.item, 0, 0, how_much)
		_tasks[task.id()] = task

		(node.graph as MatterNetworkGraph?)?.onMatterTaskCreated(task)

		listeners.forEach { menu: MatterPanelMenu -> menu.taskUpdated(task) }
		setChanged()

		return task
	}

	override fun dropAllTasks() {
		val graph = node.graph as MatterNetworkGraph?

		for (task in _tasks.values) {
			graph?.onMatterTaskRemoved(task)
			listeners.forEach { menu: MatterPanelMenu -> menu.taskUpdated(task) }
		}

		_tasks.clear()
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.matter_panel")
	}
}