package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.CapabilityItemHandler
import ru.dbotthepony.mc.otm.block.entity.worker.MatteryWorkerBlockEntity
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerJob
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerJobStatus
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerTickContext
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler
import ru.dbotthepony.mc.otm.capability.matter.MatterDirection
import ru.dbotthepony.mc.otm.capability.matter.MatterHandlerImpl
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.MatteryContainerFilter
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.item.MatterDustItem
import ru.dbotthepony.mc.otm.menu.MatterRecyclerMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.set

class MatterRecyclerBlockEntity(blockPos: BlockPos, blockState: BlockState)
	: MatteryWorkerBlockEntity(MBlockEntities.MATTER_RECYCLER, blockPos, blockState), IMatterGraphNode {
	val matter = MatterHandlerImpl(
		this::setChangedLight,
		MatterDirection.EXTRACT,
		STORAGE
	)
	val container = MatteryContainer(this::setChangedLight, 1)
	private val node = Graph6Node<IMatterGraphNode>(this)
	private var resolverNode = LazyOptional.of { this }
	private var valid = true
	override val energy = WorkerEnergyStorage(this, MAX_POWER)

	override fun getAsMatterNode(): Graph6Node<IMatterGraphNode> {
		return node
	}

	override fun getMatterHandler(): IMatterHandler {
		return matter
	}

	private val itemHandler = container.handler(object : MatteryContainerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.item is MatterDustItem
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return false
		}
	})

	override fun invalidateCaps() {
		super.invalidateCaps()
		matter.invalidate()
		itemHandler.invalidate()
		resolverNode.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		matter.revive()
		itemHandler.revive()
		resolverNode = LazyOptional.of { this }
	}

	override fun setRemoved() {
		super.setRemoved()
		node.destroy(::MatterNetworkGraph)
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			MatterNetworkGraph.discoverFull(this, node)
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["matter"] = matter.serializeNBT()
		nbt["container"] = container.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		nbt.ifHas("matter", CompoundTag::class.java, matter::deserializeNBT)
		container.deserializeNBT(nbt["container"])
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (!valid)
			return super.getCapability(cap, side)

		if (cap === CapabilityItemHandler.ITEM_HANDLER_CAPABILITY) {
			return itemHandler.get().cast()
		} else if (cap === MatteryCapability.MATTER) {
			return matter.get().cast()
		} else if (cap === MatteryCapability.MATTER_NODE) {
			return resolverNode.cast()
		}

		return super.getCapability(cap, side)
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterRecyclerMenu(containerID, inventory, this)
	}

	override fun onJobFinish(job: WorkerJob): WorkerJobStatus {
		// вся логика в onWorkTick
		return WorkerJobStatus()
	}

	override fun computeNextJob(): WorkerJob? {
		if (matter.missingMatter.isZero)
			return null

		val stack = container[0]

		if (stack.isEmpty || stack.item !is MatterDustItem) {
			return null
		}

		val copy = stack.copy()
		copy.count = 1

		val dustMatter = (stack.item as MatterDustItem).getMatterValue(copy) ?: return null
		stack.shrink(1)
		return WorkerJob(copy, dustMatter.value.toDouble() * MATTER_TICKS, POWER_CONSUMPTION)
	}

	override fun onWorkTick(context: WorkerTickContext): WorkerJobStatus {
		if ((level?.random?.nextDouble() ?: 1.0) <= 0.4)
			return WorkerJobStatus()

		val receive = if (context.ticksAdvanced == 1.0) MATTER_PER_TICK else MATTER_PER_TICK * context.ticksAdvanced
		val received = matter.receiveMatterInner(receive, true)

		if (receive != received)
			return WorkerJobStatus(false, 20)

		matter.receiveMatterInner(receive, false)
		return WorkerJobStatus()
	}

	fun tick() {
		basicTicker()

		val graph = node.graph as MatterNetworkGraph? ?: return
		val received = graph.receiveMatter(matter.storedMatter, false)

		if (!received.isZero) {
			matter.extractMatterInner(received, false)
		}
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.matter_recycler")
		private val STORAGE = ImpreciseFraction(40)
		private val POWER_CONSUMPTION = ImpreciseFraction(80)
		private val MAX_POWER = ImpreciseFraction(80_000)
		private const val MATTER_TICKS = 200.0
		private val MATTER_PER_TICK = ImpreciseFraction(1.0 / MATTER_TICKS)
	}
}
