package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.CapabilityItemHandler
import ru.dbotthepony.mc.otm.block.entity.worker.MatteryWorkerBlockEntity
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerJob
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerJobStatus
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.PatternState
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.matter.*
import ru.dbotthepony.mc.otm.menu.MatterScannerMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import java.util.*

class MatterScannerBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) :
	MatteryWorkerBlockEntity(MBlockEntities.MATTER_SCANNER, p_155229_, p_155230_), IMatterGraphNode {

	val container = MatteryContainer(this::setChanged, 1)
	override val energy = WorkerEnergyStorage(this, STORAGE, MAX_IO)
	private val itemHandler = container.handler(
		{ _: Int, stack: ItemStack -> canDecompose(stack) },
		{ _: Int, _: Int, _: ItemStack -> isIdling }
	)

	// IMatterGraphNode
	override fun getAsMatterNode(): Graph6Node<IMatterGraphNode> {
		return matterNode
	}

	override fun onPatternAdded(state: PatternState) {
		isIdling = false
	}

	override fun onPatternRemoved(state: PatternState) {
		isIdling = false
	}

	override fun onPatternUpdated(new_state: PatternState, old_state: PatternState) {
		isIdling = false
	}
	// /IMatterGraphNode

	private var valid = true
	private val matterNode: Graph6Node<IMatterGraphNode> = Graph6Node(this)
	private var resolverNode = LazyOptional.of { this }

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid) {
			if (cap === CapabilityItemHandler.ITEM_HANDLER_CAPABILITY) return itemHandler.get().cast()
			if (cap == MatteryCapability.MATTER_NODE) return resolverNode.cast()
		}

		return super.getCapability(cap, side)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		itemHandler.invalidate()
		resolverNode.invalidate()
		matterNode.destroy(::MatterNetworkGraph)
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		itemHandler.revive()
		resolverNode = LazyOptional.of { this }
	}

	override fun setRemoved() {
		super.setRemoved()
		matterNode.destroy(::MatterNetworkGraph)
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterScannerMenu(containerID, inventory, this)
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt.put("container", container.serializeNBT())
	}

	override fun load(nbt: CompoundTag) {
		container.deserializeNBT(nbt["container"])
		super.load(nbt)
	}

	override fun onJobFinish(job: WorkerJob): WorkerJobStatus {
		val grid = matterNode.graph as MatterNetworkGraph? ?: return WorkerJobStatus(false, 100)

		val stack = job.stack
		if (stack.isEmpty || !hasMatterValue(stack)) return WorkerJobStatus()

		val getState = grid.findPatterns(stack.item)
		var findState: PatternState? = null

		for (state in getState) {
			if (state.item() === stack.item) {
				if (findState == null && state.research() < 1.0) {
					findState = state
				} else if (findState != null && findState.research() < state.research()) {
					findState = state
				}
			}
		}

		val new =
			if (findState != null) {
				PatternState(findState.id(), stack.item, findState.research() + 0.2)
			} else {
				PatternState(UUID.randomUUID(), stack.item, 0.2)
			}

		if (!grid.insertPattern(new, onlyUpdate = false, simulate = false).isFailed) {
			return WorkerJobStatus()
		} else {
			return WorkerJobStatus(false, 200)
		}
	}

	override fun computeNextJob(): WorkerJob? {
		val grid = matterNode.graph as MatterNetworkGraph? ?: return null

		val stack = container.getItem(0)
		if (stack.isEmpty || !canDecompose(stack)) return null

		val getState = grid.findPatterns(stack.item)
		var findState: PatternState? = null

		for (state in getState) {
			if (state.item === stack.item && state.research < 1.0) {
				findState = state
			} else if (state.item === stack.item && state.research >= 1.0) {
				return null
			}
		}

		val new: PatternState =
			if (findState != null) {
				PatternState(findState.id, stack.item, findState.research + 0.2)
			} else {
				PatternState(UUID.randomUUID(), stack.item, 0.2)
			}

		if (!grid.insertPattern(new, onlyUpdate = false, simulate = true).isFailed) {
			val copy = stack.copy()
			copy.count = 1
			stack.shrink(1)
			container.setChanged()
			return WorkerJob(copy, getMatterValue(copy).complexity * baselineComplexityScanTicks, BASE_CONSUMPTION)
		}

		return null
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel) {
			MatterNetworkGraph.discoverFull(this, matterNode)
		}
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.matter_scanner")
		private val BASE_CONSUMPTION = ImpreciseFraction("40")
		private val STORAGE = ImpreciseFraction(80_000)
		private val MAX_IO = ImpreciseFraction(400)
	}
}