package ru.dbotthepony.mc.otm.block.entity

import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.longs.Long2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectAVLTreeMap
import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.CapabilityItemHandler
import net.minecraftforge.items.IItemHandler
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.block.RotatableMatteryBlock
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.plus
import ru.dbotthepony.mc.otm.graph.storage.BasicStorageGraphNode
import ru.dbotthepony.mc.otm.graph.storage.StorageNetworkGraph
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MNames
import ru.dbotthepony.mc.otm.storage.*
import java.lang.ref.WeakReference
import java.util.*
import kotlin.collections.ArrayList

private class SlotTuple(val slot: Int, val stack: ItemStack)
private class TrackedTuple(override val stack: ItemStackWrapper, override val id: UUID) : IStorageTuple<ItemStackWrapper> {
	val children = Int2ObjectAVLTreeMap<SlotTuple>()

	override fun toString(): String {
		return "TrackedTuple[$stack $id]"
	}
}

private fun Long.clamp(): Int {
	if (this > Int.MAX_VALUE) {
		return Int.MAX_VALUE
	}

	return this.toInt()
}

private class ItemHandlerComponent(private val parent: IItemHandler) : IStorageComponent<ItemStackWrapper> {
	override val storageType: StorageStackType<ItemStackWrapper>
		get() = OverdriveThatMatters.INSTANCE.ITEM_STORAGE()

	private val listeners = ArrayList<IStorageListener<ItemStackWrapper>>()

	override fun addListener(listener: IStorageListener<ItemStackWrapper>): Boolean {
		if (!listeners.contains(listener)) {
			listeners.add(listener)
			return true
		}

		return false
	}

	override fun removeListener(listener: IStorageListener<ItemStackWrapper>): Boolean {
		return listeners.remove(listener)
	}

	private var scanned = arrayOfNulls<ItemStack>(0)
	private var scannedMap = arrayOfNulls<TrackedTuple>(0)
	private val partitioned = Long2ObjectAVLTreeMap<ArrayList<TrackedTuple>>()
	private val index = Object2ObjectAVLTreeMap<UUID, TrackedTuple>()

	private fun removeTracked(slot: Int) {
		scanned[slot] = null
		val scannedMap = checkNotNull(scannedMap[slot]) { "Not tracking slot $slot" }
		this.scannedMap[slot] = null

		val item = scannedMap.children[slot] ?: throw IllegalStateException("${scannedMap.id} does not track $slot")
		scannedMap.children.remove(slot)
		val count = scannedMap.stack.count
		scannedMap.stack.count -= item.stack.count

		if (scannedMap.stack.count.isPositive) {
			for (listener in listeners) {
				listener.changeStack(scannedMap, count)
			}
		} else {
			for (listener in listeners) {
				listener.removeStack(scannedMap)
			}

			index.remove(scannedMap.id)

			val listing = partitioned[scannedMap.stack.longHashCode] ?: throw IllegalStateException("Item listing is not present for ${scannedMap.stack}")

			if (listing.remove(scannedMap)) {
				if (listing.isEmpty()) {
					partitioned.remove(scannedMap.stack.longHashCode)
				}
			} else {
				throw IllegalStateException("Item listing for ${scannedMap.stack} did not contain $scannedMap")
			}
		}
	}

	private fun diffTracked(slot: Int, diff: Int) {
		val scannedMap = checkNotNull(scannedMap[slot]) { "Not tracking slot $slot" }
		val item = scannedMap.children[slot] ?: throw IllegalStateException("${scannedMap.id} does not track $slot")

		val oldCount = scannedMap.stack.count
		item.stack.count += diff
		scannedMap.stack.count += diff

		for (listener in listeners) {
			listener.changeStack(scannedMap.stack, scannedMap.id, oldCount)
		}
	}

	private fun addTracked(slot: Int, stack: ItemStack) {
		check(scannedMap[slot] == null) { "Already tracking slot $slot" }

		val listing = partitioned.computeIfAbsent(ItemStackWrapper.longHashCode(stack.item), Long2ObjectFunction { ArrayList() })
		var tuple: TrackedTuple? = null

		for (storedTuple in listing) {
			if (storedTuple.stack.sameItem(stack)) {
				tuple = storedTuple
				break
			}
		}

		val added = tuple == null
		var oldCount = ImpreciseFraction.ZERO

		if (added) {
			val storageStack = ItemStackWrapper(stack)
			tuple = TrackedTuple(storageStack, UUID.randomUUID())
			index[tuple.id] = tuple
			listing.add(tuple)
		} else {
			oldCount = tuple!!.stack.count
			tuple.stack.count += stack.count
		}

		tuple.children[slot] = SlotTuple(slot, stack.copy())
		scanned[slot] = tuple.children[slot].stack
		scannedMap[slot] = tuple

		if (added) {
			for (listener in listeners) {
				listener.addStack(tuple.stack, tuple.id, this)
			}
		} else {
			for (listener in listeners) {
				listener.changeStack(tuple.stack, tuple.id, oldCount)
			}
		}
	}

	private fun sizeScan() {
		if (scanned.size != parent.slots) {
			val old = scanned
			val oldMap = scannedMap

			if (scanned.size < parent.slots) {
				// grow
				scanned = arrayOfNulls(parent.slots)
				scannedMap = arrayOfNulls(parent.slots)

				for ((i, item) in old.withIndex()) {
					scanned[i] = item
				}

				for ((i, item) in oldMap.withIndex()) {
					scannedMap[i] = item
				}
			} else {
				// shrink
				for (i in parent.slots until scanned.size) {
					removeTracked(i)
				}

				scanned = arrayOfNulls(parent.slots)
				scannedMap = arrayOfNulls(parent.slots)

				for (i in 0 until parent.slots) {
					scanned[i] = old[i]
					scannedMap[i] = oldMap[i]
				}
			}
		}
	}

	fun scan(slot: Int) {
		val current = parent[slot].let { if (it.isEmpty) null else it }
		val last = scanned[slot]

		if (current == null && last != null) {
			removeTracked(slot)
		} else if (current != null && last == null) {
			addTracked(slot, current)
		} else if (current != null && last != null) {
			if (!ItemStack.isSameItemSameTags(current, last)) {
				removeTracked(slot)
				addTracked(slot, current)
			} else if (current.count != last.count) {
				diffTracked(slot, current.count - last.count)
			}
		}
	}

	fun scan() {
		sizeScan()

		for (slot in 0 until parent.slots) {
			scan(slot)
		}
	}

	override fun insertStack(stack: ItemStackWrapper, simulate: Boolean): ItemStackWrapper {
		var leftover = stack.copy()

		for (slot in 0 until parent.slots) {
			val oldCount = leftover.count
			leftover = ItemStackWrapper(parent.insertItem(slot, leftover.stack, simulate))

			if (oldCount != leftover.count && !simulate) {
				scan(slot)
			}

			if (leftover.isEmpty) {
				break
			}
		}

		return leftover
	}

	override fun getStack(id: UUID): ItemStackWrapper {
		return index[id]?.stack ?: ItemStackWrapper.EMPTY
	}

	override fun extractStack(id: UUID, amount: ImpreciseFraction, simulate: Boolean): ItemStackWrapper {
		@Suppress("NAME_SHADOWING")
		val amount = amount.floor()

		if (!amount.isPositive)
			return ItemStackWrapper.EMPTY

		val intAmount = amount.toLong()
		val tuple = index[id] ?: return ItemStackWrapper.EMPTY
		var totalExtracted = 0L

		val iter = tuple.children.values.iterator()
		val listCopy = Array(tuple.children.values.size) {
			iter.next()
		}

		val copy = tuple.stack.copy()

		for (stack in listCopy) {
			val extracted = parent.extractItem(stack.slot, stack.stack.count.coerceAtMost((intAmount - totalExtracted).clamp()), true)

			if (extracted.isEmpty) {
				// dummy condition
				continue
			} else if (tuple.stack.sameItem(extracted)) {
				if (!simulate) {
					parent.extractItem(stack.slot, stack.stack.count.coerceAtMost((intAmount - totalExtracted).clamp()), false)
				}

				totalExtracted += extracted.count

				if (extracted.count != 0 && !simulate) {
					scan(stack.slot)
				}

				if (totalExtracted >= intAmount) {
					break
				}
			} else {
				// O SHI~
				scan(stack.slot)
			}
		}

		if (totalExtracted == 0L) {
			return ItemStackWrapper.EMPTY
		}

		copy.count = ImpreciseFraction(totalExtracted)
		return copy
	}

	override fun getStacks(): Collection<IStorageTuple<ItemStackWrapper>> {
		val listing = ArrayList<IStorageTuple<ItemStackWrapper>>(index.size)

		for (tuple in index.values) {
			listing.add(StorageTuple(tuple.id, tuple.stack))
		}

		return listing
	}
}

class StorageBusBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryPoweredBlockEntity(MBlockEntities.STORAGE_BUS, blockPos, blockState) {
	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu? {
		return null
	}

	override val energy = WorkerEnergyStorage(this)

	val cell = BasicStorageGraphNode()

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel) {
			StorageNetworkGraph.discoverFull(this, cell.storageNode)
		}

		tickOnceServer(this::checkSurroundings)
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		return if (cap === MatteryCapability.STORAGE_NODE) {
			cell.get().cast()
		} else super.getCapability(cap, side)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		cell.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		cell.revive()
	}

	private var neighbour: LazyOptional<IItemHandler>? = null
	private var component: ItemHandlerComponent? = null

	fun tick() {
		component?.scan()
	}

	override fun setRemoved() {
		super.setRemoved()
		cell.destroy(level)
	}

	fun checkSurroundings() {
		if (isRemoved)
			return

		val front = blockPos + blockState.getValue(RotatableMatteryBlock.FACING_FULL)
		val storage = level?.getBlockEntity(front)?.getCapability(CapabilityItemHandler.ITEM_HANDLER_CAPABILITY)?.let { if (it.isPresent) it else null }

		if (neighbour != storage) {
			neighbour = storage

			if (storage != null) {
				val ref = WeakReference(this)

				storage.addListener {
					val self = ref.get() ?: return@addListener

					if (self.neighbour === it && !SERVER_IS_DYING) {
						self.checkSurroundings()
					}
				}

				component?.let(cell::removeStorageComponent)
				component = ItemHandlerComponent(storage.orThrow())
				component!!.let(cell::addStorageComponent)
			} else {
				component?.let(cell::removeStorageComponent)
				component = null
			}
		}
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.${OverdriveThatMatters.MOD_ID}.${MNames.STORAGE_BUS}")
	}
}
