package ru.dbotthepony.mc.otm.block.entity.worker

import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.core.BlockPos
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.block.entity.MatteryPoweredBlockEntity
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.DoubleTag
import net.minecraft.world.level.block.Block
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.weakGreaterOrEqual
import ru.dbotthepony.mc.otm.core.weakLessThan
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.set

abstract class MatteryWorkerBlockEntity(p_155228_: BlockEntityType<*>, p_155229_: BlockPos, p_155230_: BlockState) :
	MatteryPoweredBlockEntity(p_155228_, p_155229_, p_155230_) {
	var workTicks = 0.0
		protected set

	var throttleTicks = 0
		protected set

	var currentJob: WorkerJob? = null
		protected set

	// если isIdling То ничего не делать
	// isIdling должна быть выставлена в true если что то изменилось, что могло создать работу
	var isIdling = false
		protected set

	val isUnableToProcess: Boolean get() = throttleTicks > 0

	val workProgress: Float
		get() {
			val currentJob = currentJob ?: return 0.0f
			return (workTicks / currentJob.ticks).coerceAtMost(1.0).toFloat()
		}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["work_ticks"] = workTicks
		currentJob?.let { nbt["current_job"] = it.serializeNBT() }
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas("work_ticks", DoubleTag::class.java) {
			workTicks = it.asDouble
		}

		currentJob = WorkerJob.deserializeNBT(nbt["current_job"])

		if (currentJob == null)
			workTicks = 0.0
	}

	override fun setChanged() {
		super.setChanged()
		isIdling = false
	}

	override fun setChangedLight() {
		super.setChangedLight()
		isIdling = false
	}

	/**
	 * @param job current job
	 * @return whenever machine can finish it's job. return false if machine for whatever reason can't finish it's job,
	 * waiting on conditions to be met
	 */
	protected abstract fun onJobFinish(job: WorkerJob): WorkerJobStatus

	/**
	 * @param context context for current job
	 * @return whenever machine can perform it
	 */
	protected open fun onWorkTick(context: WorkerTickContext): WorkerJobStatus {
		return WorkerJobStatus()
	}

	private var idleTicksAnim = 0
	private var workingTicksAnim = 0
	private var errorTicksAnim = 0

	override fun redstoneStatusUpdated(new_blocked: Boolean, old_blocked: Boolean) {
		super.redstoneStatusUpdated(new_blocked, old_blocked)
		isIdling = new_blocked
	}

	protected fun workerLoop() {
		if (errorTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.ERROR) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.ERROR), Block.UPDATE_CLIENTS)
		}

		if (throttleTicks > 0) {
			workingTicksAnim = 0
			idleTicksAnim = 0
			throttleTicks--
			errorTicksAnim++

			if (throttleTicks > 0)
				return
		}

		if (isIdling) {
			workingTicksAnim = 0
			errorTicksAnim = 0
			idleTicksAnim++

			if (idleTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.IDLE) {
				level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
			}

			return
		}

		var currentJob = currentJob

		if (currentJob == null) {
			if (isBlockedByRedstone) {
				isIdling = true
				return
			}

			val input = computeNextJob()

			if (input == null) {
				isIdling = true
				workingTicksAnim = 0
				return
			}

			this.currentJob = input
			currentJob = input
		}

		if (isBlockedByRedstone) {
			isIdling = true
			return
		}

		if (!currentJob.power.isZero && energy.batteryLevel.isZero) {
			idleTicksAnim++

			if (idleTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.IDLE) {
				level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
			}

			return
		}

		idleTicksAnim = 0

		if (weakLessThan(workTicks, currentJob.ticks)) {
			if (!currentJob.power.isZero) {
				// сколько осталось тиков работать
				val ticksLeft = currentJob.ticks - workTicks
				val requiredPower: ImpreciseFraction

				// запрос энергии на то количество, сколько действительно осталось работать
				if (ticksLeft > 1.0) {
					requiredPower = currentJob.power
				} else {
					requiredPower = currentJob.power * ticksLeft
				}

				val extractedPower = if (requiredPower.isZero) ImpreciseFraction.ZERO else energy.extractEnergyInner(requiredPower, true)

				// сколько тиков мы "проработали"
				// может быть меньше, чем единица, если недостаточно питания или мы завершаем работу,
				// для которой осталось дробное количество тиков
				val ticksAdvanced = (extractedPower / requiredPower).toDouble().coerceAtMost(ticksLeft)
				val ticksAdvancedWeak = if (requiredPower.isZero) 1.0 else ticksAdvanced
				val status = onWorkTick(WorkerTickContext(currentJob, requiredPower, extractedPower, ticksAdvanced))

				if (!status.valid) {
					throttleTicks += status.throttle
					return
				}

				workingTicksAnim++
				errorTicksAnim = 0

				val updatedWorkTicks = (workTicks + ticksAdvancedWeak).coerceAtMost(currentJob.ticks)

				if (weakGreaterOrEqual(updatedWorkTicks, currentJob.ticks)) {
					workTicks = currentJob.ticks
					energy.extractEnergyInner(extractedPower * (1.0 - (updatedWorkTicks - currentJob.ticks)), false)

					val finish = onJobFinish(currentJob)

					if (finish.valid) {
						this.currentJob = null
						workTicks = 0.0
					} else {
						throttleTicks += finish.throttle
					}
				} else {
					workTicks = updatedWorkTicks
					energy.extractEnergyInner(extractedPower, false)
				}
			} else {
				val ticksLeft = (currentJob.ticks - workTicks).coerceAtMost(1.0)
				val status = onWorkTick(WorkerTickContext(currentJob, ImpreciseFraction.ZERO, ImpreciseFraction.ZERO, ticksLeft))

				if (!status.valid) {
					throttleTicks += status.throttle
					return
				}

				workingTicksAnim++
				errorTicksAnim = 0
				workTicks = (workTicks + 1.0).coerceAtMost(currentJob.ticks)

				if (weakGreaterOrEqual(workTicks, currentJob.ticks)) {
					val finish = onJobFinish(currentJob)

					if (finish.valid) {
						this.currentJob = null
						workTicks = 0.0
					} else {
						throttleTicks += finish.throttle
					}
				}
			}
		} else {
			val finish = onJobFinish(currentJob)

			if (finish.valid) {
				this.currentJob = null
				workTicks = 0.0
				errorTicksAnim = 0
			} else {
				throttleTicks += finish.throttle
				errorTicksAnim++
			}
		}

		if (workingTicksAnim > 20 &&
			errorTicksAnim == 0 &&
			blockState.hasProperty(WorkerState.WORKER_STATE) &&
			blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.WORKING)
		{
			level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS)
		}
	}

	/**
	 * Determine which item can be processed from input slots if idling
	 * @return any item in input slots. null if no work is available
	 */
	protected abstract fun computeNextJob(): WorkerJob?

	fun basicTicker() {
		batteryChargeLoop()
		workerLoop()
	}
}