package ru.dbotthepony.mc.otm.capability.matter

import net.minecraft.world.item.Item
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import java.util.*
import java.util.function.Predicate
import java.util.stream.Collectors

interface IMatterHandler {
	val storedMatter: ImpreciseFraction
	val maxStoredMatter: ImpreciseFraction

	fun receiveMatterOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction
	fun receiveMatterInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction
	fun extractMatterOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction
	fun extractMatterInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction
	val direction: MatterDirection
	val missingMatter: ImpreciseFraction
		get() = maxStoredMatter.minus(storedMatter).moreThanZero()

	val allowsExtract: Boolean
		get() = direction != MatterDirection.RECEIVE

	val allowsReceive: Boolean
		get() = direction != MatterDirection.EXTRACT
}

enum class MatterDirection { RECEIVE, EXTRACT, BIDIRECTIONAL }

interface IMatterTaskProvider {
	/**
	 * @return immutable collection of tasks that can be allocated by a worker
	 */
	val tasks: Collection<MatterTask>

	/**
	 * @return immutable collection of all stored tasks, even fully allocated by workers
	 */
	val allTasks: Collection<MatterTask>

	/**
	 * Allocates (marks as work-in-progress) a task
	 * by incrementing it's in_progress by 1
	 * and shrinking required by 1
	 *
	 * If required == 0, it should not be returned by this method
	 * @param simulate whenever to change internal state
	 * @return MatterTaskAllocation(task, pattern) that should be performed, or null if no work is available
	 */
	fun allocateTask(simulate: Boolean): MatterTaskAllocation?

	/**
	 * Notify about task completion. If this provider indeed contain this task, it should
	 * shrink in_progress by 1
	 * If in_progress == 0 and required == 0, it should discard the task
	 * @param task task being completed. this method should ignore tasks that are not owned by it.
	 * @return whenever task indeed belong to this provider and internal state was updated
	 */
	fun notifyTaskCompletion(task: MatterTask): Boolean

	/**
	 * @param id uuid of task
	 * @return MatterTask that this capability holds with this id, or null
	 */
	fun getTask(id: UUID): MatterTask?

	/**
	 * Destroys all tasks this capability contains
	 */
	fun dropAllTasks()
}

interface IPatternStorage {
	/**
	 * @return unmodifiable collection of stored patterns
	 */
	val storedPatterns: Collection<PatternState>

	fun findPatterns(item: Item): Collection<PatternState> {
		return findPatterns { item2: PatternState -> item == item2.item() }
	}

	fun findPatterns(predicate: Predicate<PatternState>?): Collection<PatternState> {
		return storedPatterns.stream().filter(predicate).collect(Collectors.toList())
	}

	fun findPattern(item: Item): PatternState? {
		return findPattern { item2: PatternState -> item == item2.item() }
	}

	fun findPattern(predicate: Predicate<PatternState>): PatternState? {
		for (pattern in storedPatterns) {
			if (predicate.test(pattern)) {
				return pattern
			}
		}

		return null
	}

	fun getPattern(id: UUID?): PatternState? {
		id ?: return null

		for (pattern in storedPatterns) {
			if (pattern.id == id) {
				return pattern
			}
		}

		return null
	}

	fun hasPattern(state: PatternState): Boolean {
		return getPattern(state.id) != null
	}

	fun hasPattern(id: UUID?): Boolean {
		return getPattern(id) != null
	}

	val capacity: Int
	val stored: Int

	/**
	 * Prefer way to call it is to first call only_update = true
	 * and if it fail, try only_update = false
	 *
	 * @param pattern pattern to be inserted or update value from
	 * @param only_update do not insert new pattern if this pattern's UUID is not matched
	 * @param simulate whenever to affect state
	 * @return record of status of the operation (at status() FAIL, UPDATED, INSERTED) as well as new_state and old_state
	 */
	fun insertPattern(pattern: PatternState, only_update: Boolean, simulate: Boolean): PatternInsertStatus

	fun insertPattern(pattern: PatternState, simulate: Boolean): PatternInsertStatus {
		return insertPattern(pattern, false, simulate)
	}

	fun updatePattern(pattern: PatternState, simulate: Boolean): PatternInsertStatus {
		return insertPattern(pattern, true, simulate)
	}
}
