package ru.dbotthepony.mc.otm.data

import com.google.common.collect.ImmutableList
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.storage.loot.Deserializers
import net.minecraft.world.level.storage.loot.LootContext
import net.minecraft.world.level.storage.loot.LootPool
import net.minecraft.world.level.storage.loot.predicates.LootItemCondition
import net.minecraftforge.common.loot.GlobalLootModifierSerializer
import net.minecraftforge.common.loot.LootModifier
import net.minecraftforge.registries.ForgeRegistries
import ru.dbotthepony.mc.otm.OverdriveThatMatters

class LootTableAppender(conditions: Array<out LootItemCondition>, private vararg val pools: LootPool) : LootModifier(conditions) {
	override fun doApply(generatedLoot: MutableList<ItemStack>, context: LootContext): MutableList<ItemStack> {
		pools.forEach { it.addRandomItems(generatedLoot::add, context) }
		return generatedLoot
	}

	companion object : GlobalLootModifierSerializer<LootTableAppender>() {
		private val lootPoolGson = Deserializers.createLootTableSerializer().create()

		override fun read(
			location: ResourceLocation,
			`object`: JsonObject,
			ailootcondition: Array<out LootItemCondition>
		): LootTableAppender {
			val pools = lootPoolGson.fromJson(`object`["pools"], Array<LootPool>::class.java) ?: throw JsonParseException("Invalid 'pools' entry")
			pools.forEach {
				@Suppress("UNNECESSARY_NOT_NULL_ASSERTION")
				it!!
			}
			return LootTableAppender(ailootcondition, *pools)
		}

		override fun write(instance: LootTableAppender): JsonObject {
			val build = makeConditions(instance.conditions)
			build.add("pools", lootPoolGson.toJsonTree(instance.pools))
			return build
		}

		init {
			registryName = ResourceLocation(OverdriveThatMatters.MOD_ID, "loot_appender")
		}
	}
}

class LootTableBasicAppender(conditions: Array<out LootItemCondition>, items: List<Pair<ItemStack, Double>>) : LootModifier(conditions) {
	private val items = ImmutableList.copyOf(items)

	constructor(conditions: Array<out LootItemCondition>, vararg items: Pair<ItemStack, Double>) : this(conditions, items.toList())

	override fun doApply(generatedLoot: MutableList<ItemStack>, context: LootContext): MutableList<ItemStack> {
		for ((item, chance) in items) {
			if (context.random.nextDouble() <= chance) {
				generatedLoot.add(item.copy())
			}
		}

		return generatedLoot
	}

	companion object : GlobalLootModifierSerializer<LootTableBasicAppender>() {
		override fun read(
			location: ResourceLocation,
			obj: JsonObject,
			ailootcondition: Array<out LootItemCondition>
		): LootTableBasicAppender {
			val entries = ArrayList<Pair<ItemStack, Double>>()

			for (entry in obj["entries"].asJsonArray) {
				entry as JsonObject
				val item = entry["item"]?.asJsonObject ?: throw JsonParseException("Missing 'item'")
				val name = item["name"]?.asString ?: throw JsonParseException("Missing 'item.name'")
				val count = item["count"]?.asInt ?: throw JsonParseException("Missing 'item.count'")
				val chance = entry["chance"]?.asDouble ?: throw JsonParseException("Missing 'chance'")

				val itemObj = ForgeRegistries.ITEMS.getValue(ResourceLocation(name)) ?: throw JsonParseException("Unknown item $name")
				val stack = ItemStack(itemObj, count)

				entries.add(stack to chance)
			}

			return LootTableBasicAppender(ailootcondition, ImmutableList.copyOf(entries))
		}

		override fun write(instance: LootTableBasicAppender): JsonObject {
			val obj = makeConditions(instance.conditions)
			val listing = JsonArray()

			for ((item, chance) in instance.items) {
				val entry = JsonObject()

				entry.add("item", JsonObject().also {
					it.addProperty("name", item.item.registryName!!.toString())
					it.addProperty("count", item.count)
				})

				entry.addProperty("chance", chance)

				listing.add(entry)
			}

			obj.add("entries", listing)

			return obj
		}

		init {
			registryName = ResourceLocation(OverdriveThatMatters.MOD_ID, "loot_appender_basic")
		}
	}
}
