package ru.dbotthepony.mc.otm.graph

import it.unimi.dsi.fastutil.objects.Object2ObjectAVLTreeMap
import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.core.SectionPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.level.block.entity.BlockEntity
import ru.dbotthepony.mc.otm.IConditionalTickable
import ru.dbotthepony.mc.otm.addPreWorldTicker
import ru.dbotthepony.mc.otm.core.plus
import java.util.*
import kotlin.collections.ArrayList

abstract class Abstract6Graph<T> : IConditionalTickable {
	@JvmField
	protected val _nodes = ArrayList<Graph6Node<T>>()
	fun size() = _nodes.size

	private val tickable = ArrayList<Graph6Node<T>>()

	@JvmField
	val nodes: Collection<Graph6Node<T>> = Collections.unmodifiableCollection(_nodes)

	/**
	 * Allows storing arbitrary data by external code
	 */
	@JvmField
	val userData = Object2ObjectAVLTreeMap<UUID, Any>()

	abstract fun onNodeRemoved(node: Graph6Node<T>)
	abstract fun onNodeAdded(node: Graph6Node<T>)

	override val canTick: Boolean
		get() = _nodes.size > 0

	override fun tick() {
		for (i in tickable.size - 1 downTo 0) {
			val node = tickable[i]

			if (node.canTick) {
				node.tick()
			} else {
				tickable.removeAt(i)
			}
		}
	}

	fun removeNode(node: Graph6Node<T>) {
		if (!_nodes.contains(node))
			throw IllegalStateException("Not containing node $node")

		_nodes.remove(node)
		node.graph = null
		onNodeRemoved(node)

		if (node.canTick) {
			tickable.remove(node)
		}
	}

	fun addNode(node: Graph6Node<T>) {
		if (_nodes.contains(node))
			throw IllegalStateException("Already containing node $node")

		_nodes.add(node)
		node.graph = this
		onNodeAdded(node)

		if (node.canTick) {
			tickable.add(node)
		}
	}

	fun merge(other: Abstract6Graph<T>): Abstract6Graph<T> {
		if (other === this)
			return this

		if (size() >= other.size()) {
			for (node in other._nodes) {
				_nodes.add(node)
				node.graph = this
				onNodeAdded(node)
			}

			return this
		} else {
			return other.merge(this)
		}
	}

	companion object {
		fun <T> discoverFull(
			level: ServerLevel,
			blockPos: BlockPos,
			node: Graph6Node<T>,
			nodeGetter: (BlockEntity) -> Graph6Node<T>?,
			factory: () -> Abstract6Graph<T>
		) {
			addPreWorldTicker(level, object : IConditionalTickable {
				override fun tick() {
					discovered = discover(level, blockPos, node, nodeGetter, factory)
				}

				private var discovered = false

				override val canTick: Boolean
					get() = !discovered && node.valid
			})
		}

		@JvmStatic
		fun <T> discover(
			level: ServerLevel,
			blockPos: BlockPos,
			node: Graph6Node<T>,
			nodeGetter: (BlockEntity) -> Graph6Node<T>?,
			factory: () -> Abstract6Graph<T>
		): Boolean {
			var fullDiscovery = true

			node.nullifyConnections()

			var _graph = node.graph

			// для начала найдем граф к которому будем принадлежать
			if (_graph == null) {
				for (dir in Direction.values()) {
					val offset = blockPos + dir
					val chunk = level.chunkSource.getChunkNow(SectionPos.blockToSectionCoord(offset.x), SectionPos.blockToSectionCoord(offset.z))

					if (chunk == null) {
						fullDiscovery = false
						continue
					}

					val entity = chunk.getBlockEntity(offset)

					if (entity != null) {
						val getNode = nodeGetter(entity)

						if (getNode?.graph != null) {
							_graph = getNode.graph
							break
						}
					}
				}

				// мы нашли граф рядом
				if (_graph != null) {
					node.graph = _graph
					_graph.addNode(node)
				} else {
					// графов рядом нет, создаем свой
					_graph = factory()
					node.graph = _graph
					_graph.addNode(node)
				}
			}

			// теперь снова смотрим на соседей, если у них нет графа - присоединяем к своему
			// если у них есть граф - слияем его со своим или свой с его
			for (dir in Direction.values()) {
				val offset = blockPos + dir
				val chunk = level.chunkSource.getChunkNow(SectionPos.blockToSectionCoord(offset.x), SectionPos.blockToSectionCoord(offset.z))

				if (chunk == null) {
					fullDiscovery = false
					continue
				}

				val entity = chunk.getBlockEntity(offset)

				if (entity != null) {
					val getNode = nodeGetter(entity)

					if (getNode != null) {
						// у вершины нет своего графа
						// добавляем в свой граф
						if (getNode.graph == null) {
							getNode.graph = node.graph!!
							node.graph!!.addNode(getNode)
						} else if (getNode.graph != node.graph) {
							// у вершины уже есть свой граф, и он не наш
							// произведём слияние графов
							val merged = getNode.graph!!.merge(node.graph!!)
							getNode.graph = merged
							node.graph = merged
						}

						node.setToNeightbour(getNode, dir)
					}
				}
			}

			return fullDiscovery
		}
	}
}
