package ru.dbotthepony.mc.otm.graph.matter

import net.minecraft.server.level.ServerLevel
import net.minecraft.world.item.Item
import net.minecraft.world.level.block.entity.BlockEntity
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.matter.*
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.graph.Abstract6Graph
import ru.dbotthepony.mc.otm.graph.Graph6Node
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashSet

@Suppress("unused")
class MatterNetworkGraph : Abstract6Graph<IMatterGraphNode>(), IMatterGraphListener {
	private val listeners = HashSet<IMatterGraphListener>()

	fun addListener(listener: IMatterGraphListener) = listeners.add(listener)
	fun removeListener(listener: IMatterGraphListener) = listeners.remove(listener)

	override fun onNodeRemoved(node: Graph6Node<IMatterGraphNode>) {
		val patterns = node.value.getPatternHandler()

		if (patterns != null) {
			for (pattern in patterns.storedPatterns) {
				onPatternRemoved(pattern)
			}
		}

		val tasks = node.value.getTaskHandler()

		if (tasks != null) {
			for (task in tasks.tasks) {
				onMatterTaskRemoved(task)
			}
		}

		for (pattern in getStoredPatterns()) {
			node.value.onPatternRemoved(pattern)
		}

		for (task in getTasks()) {
			node.value.onMatterTaskRemoved(task)
		}
	}

	override fun onNodeAdded(node: Graph6Node<IMatterGraphNode>) {
		for (pattern in getStoredPatterns()) {
			node.value.onPatternAdded(pattern)
		}

		for (task in getTasks()) {
			node.value.onMatterTaskCreated(task)
		}

		val patterns = node.value.getPatternHandler()

		if (patterns != null) {
			for (pattern in patterns.storedPatterns) {
				onPatternAdded(pattern)
			}
		}

		val tasks = node.value.getTaskHandler()

		if (tasks != null) {
			for (task in tasks.tasks) {
				onMatterTaskCreated(task)
			}
		}
	}

	fun getMatterStorageLevel(): ImpreciseFraction {
		var level = ImpreciseFraction.ZERO

		for (node in _nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null && matter.direction == MatterDirection.BIDIRECTIONAL) {
				level += matter.storedMatter
			}
		}

		return level
	}

	fun getMatterStorageMaxLevel(): ImpreciseFraction {
		var level = ImpreciseFraction.ZERO

		for (node in _nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null && matter.direction == MatterDirection.BIDIRECTIONAL) {
				level += matter.maxStoredMatter
			}
		}

		return level
	}

	fun extractMatter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (howMuch <= ImpreciseFraction.ZERO)
			return ImpreciseFraction.ZERO

		var howMuch = howMuch
		var extracted = ImpreciseFraction.ZERO

		for (node in _nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null) {
				val value = matter.extractMatterOuter(howMuch, simulate)
				howMuch -= value
				extracted += value

				if (howMuch <= ImpreciseFraction.ZERO)
					return extracted
			}
		}

		return extracted
	}

	fun receiveMatter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (howMuch <= ImpreciseFraction.ZERO)
			return ImpreciseFraction.ZERO

		var howMuch = howMuch
		var received = ImpreciseFraction.ZERO

		for (node in _nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null && matter.direction == MatterDirection.BIDIRECTIONAL) {
				val value = matter.receiveMatterOuter(howMuch, simulate)
				howMuch -= value
				received += value

				if (howMuch <= ImpreciseFraction.ZERO)
					return received
			}
		}

		return received
	}

	fun receiveMatterForce(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (howMuch <= ImpreciseFraction.ZERO)
			return ImpreciseFraction.ZERO

		var howMuch = howMuch
		var received = ImpreciseFraction.ZERO

		for (node in _nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null && matter.direction != MatterDirection.EXTRACT) {
				val value = matter.receiveMatterOuter(howMuch, simulate)
				howMuch -= value
				received += value

				if (howMuch <= ImpreciseFraction.ZERO)
					return received
			}
		}

		return received
	}

	private fun _insertPattern(state: PatternState, onlyUpdate: Boolean, simulate: Boolean): PatternInsertStatus {
		for (node in _nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				val status = storage.insertPattern(state, onlyUpdate, simulate)

				if (!status.isFailed) {
					return status
				}
			}
		}

		return PatternInsertFailure
	}

	fun insertPattern(state: PatternState, onlyUpdate: Boolean, simulate: Boolean): PatternInsertStatus {
		if (onlyUpdate) return _insertPattern(state, true, simulate)
		val status = _insertPattern(state, true, simulate)
		if (!status.isFailed) return status
		return _insertPattern(state, false, simulate)
	}

	fun getTasks(): Collection<MatterTask> {
		val list = ArrayList<MatterTask>()

		for (node in _nodes) {
			val tasks = node.value.getTaskHandler()

			if (tasks != null) {
				list.addAll(tasks.tasks)
			}
		}

		return list
	}

	fun getStoredPatterns(): Collection<PatternState> {
		val list = ArrayList<PatternState>()

		for (node in _nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				list.addAll(storage.storedPatterns)
			}
		}

		return list
	}

	fun getPatternCount(): Long {
		var value = 0L

		for (node in _nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				value += storage.stored
			}
		}

		return value
	}

	fun getPatternCapacity(): Long {
		var value = 0L

		for (node in _nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				value += storage.capacity
			}
		}

		return value
	}

	fun getPattern(id: UUID): PatternState? {
		for (node in _nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				val get = storage.getPattern(id)

				if (get != null) {
					return get
				}
			}
		}

		return null
	}

	fun getPattern(state: PatternState) = getPattern(state.id)
	fun hasPattern(state: PatternState) = getPattern(state.id) != null
	fun hasPattern(id: UUID) = getPattern(id) != null

	fun findPattern(predicate: (PatternState) -> Boolean): PatternState? {
		for (node in _nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				for (state in storage.storedPatterns) {
					if (predicate(state)) {
						return state
					}
				}
			}
		}

		return null
	}

	fun findPattern(predicate: Item) = findPattern { it.item == predicate }

	fun findPatterns(predicate: (PatternState) -> Boolean): Collection<PatternState> {
		val list = ArrayList<PatternState>()

		for (node in _nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				for (state in storage.storedPatterns) {
					if (predicate(state)) {
						list.add(state)
					}
				}
			}
		}

		return list
	}

	fun findPatterns(predicate: Item) = findPatterns { it.item == predicate }

	fun allocateTask(simulate: Boolean): MatterTaskAllocation? {
		for (node in _nodes) {
			val tasks = node.value.getTaskHandler()

			if (tasks != null) {
				val allocated = tasks.allocateTask(simulate)

				if (allocated != null) {
					return allocated
				}
			}
		}

		return null
	}

	fun notifyTaskCompletion(task: MatterTask): Boolean {
		for (node in _nodes) {
			val tasks = node.value.getTaskHandler()

			if (tasks != null && tasks.notifyTaskCompletion(task)) {
				return true
			}
		}

		return false
	}

	override fun onPatternAdded(state: PatternState) {
		for (node in _nodes) node.value.onPatternAdded(state)
		for (node in listeners) node.onPatternAdded(state)
	}

	override fun onPatternRemoved(state: PatternState) {
		for (node in _nodes) node.value.onPatternRemoved(state)
		for (node in listeners) node.onPatternRemoved(state)
	}

	override fun onPatternUpdated(new_state: PatternState, old_state: PatternState) {
		for (node in _nodes) node.value.onPatternUpdated(new_state, old_state)
		for (node in listeners) node.onPatternUpdated(new_state, old_state)
	}

	override fun onMatterTaskCreated(task: MatterTask) {
		for (node in _nodes) node.value.onMatterTaskCreated(task)
		for (node in listeners) node.onMatterTaskCreated(task)
	}

	override fun onMatterTaskUpdated(new_state: MatterTask, old_state: MatterTask) {
		for (node in _nodes) node.value.onMatterTaskUpdated(new_state, old_state)
		for (node in listeners) node.onMatterTaskUpdated(new_state, old_state)
	}

	override fun onMatterTaskFinished(state: MatterTask) {
		for (node in _nodes) node.value.onMatterTaskFinished(state)
		for (node in listeners) node.onMatterTaskFinished(state)
	}

	override fun onMatterTaskRemoved(state: MatterTask) {
		for (node in _nodes) node.value.onMatterTaskRemoved(state)
		for (node in listeners) node.onMatterTaskRemoved(state)
	}

	companion object {
		@JvmStatic
		fun discoverFull(tile: BlockEntity, node: Graph6Node<IMatterGraphNode>) {
			if (tile.level !is ServerLevel)
				return

			return discoverFull(
				tile.level!! as ServerLevel,
				tile.blockPos,
				node,
				fun(_tile): Graph6Node<IMatterGraphNode>? {
					val resolve = _tile.getCapability(MatteryCapability.MATTER_NODE)

					return if (resolve.isPresent) {
						resolve.resolve().get().getAsMatterNode()
					} else {
						null
					}
				},
				::MatterNetworkGraph
			)
		}

		@JvmStatic
		fun discover(tile: BlockEntity, node: Graph6Node<IMatterGraphNode>): Boolean {
			if (tile.level !is ServerLevel)
				return false

			return discover(
				tile.level!! as ServerLevel,
				tile.blockPos,
				node,
				fun(_tile): Graph6Node<IMatterGraphNode>? {
					val resolve = _tile.getCapability(MatteryCapability.MATTER_NODE)

					return if (resolve.isPresent) {
						resolve.resolve().get().getAsMatterNode()
					} else {
						null
					}
				},
				::MatterNetworkGraph
			)
		}
	}
}
