package ru.dbotthepony.mc.otm.item.weapon

import com.mojang.math.Vector3f
import net.minecraft.client.Minecraft
import net.minecraft.client.model.HumanoidModel
import net.minecraft.client.renderer.block.model.ItemTransforms
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Rarity
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.client.event.EntityViewRenderEvent
import net.minecraftforge.client.event.InputEvent
import net.minecraftforge.client.event.RenderHandEvent
import net.minecraftforge.client.event.RenderPlayerEvent
import net.minecraftforge.event.TickEvent
import net.minecraftforge.eventbus.api.SubscribeEvent
import net.minecraftforge.fml.LogicalSide
import net.minecraftforge.network.NetworkEvent
import ru.dbotthepony.kvector.util.linearInterpolation
import ru.dbotthepony.kvector.vector.Angle3f
import ru.dbotthepony.kvector.vector.ndouble.Vector3d
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.network.MatteryNetworking
import java.util.*
import java.util.function.Supplier
import kotlin.math.PI
import kotlin.reflect.KClass
import kotlin.reflect.full.isSubclassOf
import kotlin.reflect.full.primaryConstructor

private val ItemStack.weaponDataTable get() = tag?.let(::WeaponDataTable)

enum class WeaponScopePacket(val scope: Boolean) {
	SCOPE_IN(true), SCOPE_OUT(false);

	fun write(buff: FriendlyByteBuf) {
		buff.writeBoolean(scope)
	}

	fun play(supplier: Supplier<NetworkEvent.Context>) {
		supplier.get().packetHandled = true

		// TODO: Manual synchronization
		supplier.get().enqueueWork {
			val stack = supplier.get().sender!!.mainHandItem
			val item = stack.item as? AbstractWeaponItem<*> ?: return@enqueueWork

			item.dataTable(stack).wantsToScope = scope
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf) = if (buff.readBoolean()) SCOPE_IN else SCOPE_OUT
	}
}

enum class WeaponFireInputPacket(val primary: Boolean) {
	PRIMARY(true), SECONDARY(false);

	fun write(buff: FriendlyByteBuf) {
		buff.writeBoolean(primary)
	}

	fun play(supplier: Supplier<NetworkEvent.Context>) {
		supplier.get().packetHandled = true

		// TODO: Manual synchronization
		supplier.get().enqueueWork {
			val stack = supplier.get().sender!!.mainHandItem
			val item = stack.item as? AbstractWeaponItem<*> ?: return@enqueueWork

			// Listen server: client and server thread compete for lock
			// so it is very likely item is being in predicted context
			val predictedData = item.dataTable
			item.dataTable = null

			if (primary)
				item.tryPrimaryFire(stack, supplier.get().sender!!)
			else
				item.trySecondaryFire(stack, supplier.get().sender!!)

			(item as AbstractWeaponItem<WeaponDataTable>).dataTable = predictedData
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf) = if (buff.readBoolean()) PRIMARY else SECONDARY
	}
}

open class WeaponDataTable(val tag: CompoundTag) {
	var shotCooldown by tag.ints
	var wantsToScope by tag.booleans
	var scoped by tag.booleans
	var scopeTicks by tag.ints
	var nextPrimaryFire by tag.ints
	var nextSecondaryFire by tag.ints
	var uuid by tag.uuids
}

abstract class AbstractWeaponItem<D : WeaponDataTable>(val tables: KClass<D>, rarity: Rarity = Rarity.UNCOMMON) : Item(
	Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB).stacksTo(1).rarity(rarity)) {
	fun makeDataTable(tag: CompoundTag) = tables.primaryConstructor!!.call(tag)

	/**
	 * Predicted data table, for client use only
	 */
	var dataTable: D? = null
		set(value) {
			field = if (value == null || value::class.isSubclassOf(tables)) {
				value
			} else {
				throw ClassCastException("Reified generics: ${value::class.qualifiedName} cannot be cast to ${tables.qualifiedName}")
			}
		}

	fun compatibleDataTable(value: WeaponDataTable?) = value == null || value::class.isSubclassOf(tables)

	@Suppress("unchecked_cast")
	fun dataTable(itemStack: ItemStack): D {
		if (dataTable != null) {
			return dataTable!!
		}

		return makeDataTable(itemStack.tagNotNull)
	}

	open fun holster(itemStack: ItemStack, player: Player, dt: D = dataTable(itemStack)) {
		dt.wantsToScope = false
	}

	open fun deploy(itemStack: ItemStack, player: Player, dt: D = dataTable(itemStack)) {
		dt.wantsToScope = false

		if (dt.uuid == EMPTY_UUID) {
			dt.uuid = UUID.randomUUID()
		}
	}

	abstract val roundsPerMinute: Int
	val roundsPerSecond get() = roundsPerMinute / 60
	val fireCooldown get() = (1200 / roundsPerMinute).coerceAtLeast(1)

	open val positionIdle get() = Vector3d(1.0, -0.5, -1.0)
	open val positionIronSights get() = Vector3d(0.0, -0.23, -1.0)
	open val rotIdle get() = Angle3f(PI.toFloat() / 36f, PI.toFloat() / 18f)
	open val rotIronSights get() = Angle3f.ZERO

	open val primaryAutomatic = true
	open val secondaryAutomatic = true

	open val ironSightsFOV: Double get() = 2.0

	fun ironSightsProgress(itemStack: ItemStack, partialTicks: Double = 0.0, dt: D = dataTable(itemStack)): Double {
		if (!dt.wantsToScope)
			return (dt.scopeTicks.toDouble() - partialTicks).coerceAtLeast(0.0) / scopingTime.toDouble()

		return (dt.scopeTicks.toDouble() + partialTicks) / scopingTime.toDouble()
	}

	abstract val scopingTime: Int

	open fun think(itemStack: ItemStack, player: Player, dt: D = dataTable(itemStack)) {
		if (dt.wantsToScope && dt.scopeTicks < scopingTime) {
			dt.scopeTicks++
		} else if (!dt.wantsToScope && dt.scopeTicks > 0) {
			dt.scopeTicks--
		}
	}

	open fun thinkHolstered(itemStack: ItemStack, player: Player, dt: D = dataTable(itemStack)) {
		if (dt.shotCooldown > 0) {
			dt.shotCooldown--
		}
	}

	open fun think2(itemStack: ItemStack, player: Player, dt: D = dataTable(itemStack)) {
		if (dt.nextPrimaryFire > 0) {
			dt.nextPrimaryFire--
		}

		if (dt.nextSecondaryFire > 0) {
			dt.nextSecondaryFire--
		}
	}

	open fun primaryFire(itemStack: ItemStack, player: Player, dt: D = dataTable(itemStack)): Boolean {
		return false
	}

	open fun secondaryFire(itemStack: ItemStack, player: Player, dt: D = dataTable(itemStack)): Boolean {
		return false
	}

	fun tryPrimaryFire(itemStack: ItemStack, player: Player, dt: D = dataTable(itemStack)): Boolean {
		if (dt.nextPrimaryFire <= 0 && primaryFire(itemStack, player, dt)) {
			dt.nextPrimaryFire = fireCooldown
			return true
		}

		return false
	}

	fun trySecondaryFire(itemStack: ItemStack, player: Player, dt: D = dataTable(itemStack)): Boolean {
		if (dt.nextSecondaryFire <= 0 && secondaryFire(itemStack, player, dt)) {
			dt.nextSecondaryFire = fireCooldown
			return true
		}

		return false
	}

	override fun getDestroySpeed(p_41425_: ItemStack, p_41426_: BlockState): Float {
		return 0f
	}

	override fun onLeftClickEntity(stack: ItemStack?, player: Player?, entity: Entity?): Boolean {
		return true
	}

	companion object {
		// shared
		private val items = WeakHashMap<Player, ItemStack>()

		// client only
		private var inRender = false
		var predictedData: WeaponDataTable? = null

		@SubscribeEvent
		@Suppress("unused")
		fun fovHook(event: EntityViewRenderEvent.FieldOfView) {
			val player = event.camera.entity as? Player ?: return

			(player.mainHandItem.item as? AbstractWeaponItem<*>)?.let {
				@Suppress("unchecked_cast")
				if (it.compatibleDataTable(predictedData))
					(it as AbstractWeaponItem<WeaponDataTable>).dataTable = predictedData

				event.fov /= linearInterpolation(
					it.ironSightsProgress(
						player.mainHandItem,
						event.partialTicks
					) * 1.7 - 0.6, 1.0, it.ironSightsFOV
				)
				it.dataTable = null
			}
		}

		@SubscribeEvent
		@Suppress("unused")
		fun playerRenderHook(event: RenderPlayerEvent.Pre) {
			if (event.player.mainHandItem.item is AbstractWeaponItem<*> && event.player.offhandItem.isEmpty) {
				event.renderer.model.rightArmPose = HumanoidModel.ArmPose.BOW_AND_ARROW
				event.renderer.model.leftArmPose = HumanoidModel.ArmPose.EMPTY
			}
		}

		@SubscribeEvent
		@Suppress("unused")
		fun clickHook(event: InputEvent.ClickInputEvent) {
			val player = Minecraft.getInstance().player!!

			if (player.mainHandItem.item is AbstractWeaponItem<*> && player.offhandItem.isEmpty && event.isAttack) {
				event.isCanceled = true
				event.setSwingHand(false)
			}
		}

		@SubscribeEvent
		@Suppress("unused", "unchecked_cast")
		fun renderViewModel(event: RenderHandEvent) {
			if (inRender) {
				return
			}

			val stack = event.itemStack
			val item = stack.item as? AbstractWeaponItem<*> ?: return

			inRender = true
			event.isCanceled = true

			(item as AbstractWeaponItem<WeaponDataTable>).dataTable = predictedData

			val player = Minecraft.getInstance().player!!
			val pose = event.poseStack
			val itemInHandRenderer = Minecraft.getInstance().itemInHandRenderer

			pose.pushPose()

			val progress = item.ironSightsProgress(stack, event.partialTicks.toDouble())

			val (x, y, z) = Vector3d.bezier(
				progress,
				item.positionIdle,
				item.positionIdle,
				item.positionIdle,
				Vector3d(0.0, -1.0, -1.0),
				item.positionIronSights,
				item.positionIronSights,
				item.positionIronSights,
			)

			pose.translate(x, y, z)

			val (pitch, yaw, roll) = Angle3f.bezier(
				progress.toFloat(),
				item.rotIdle,
				item.rotIdle,
				item.rotIdle,
				item.rotIdle.copy(roll = -PI.toFloat() / 6f + item.rotIdle.roll),
				item.rotIronSights,
				item.rotIronSights,
				item.rotIronSights,
				item.rotIronSights,
				item.rotIronSights,
			)

			pose.mulPose(Vector3f.ZP.rotation(roll))
			pose.mulPose(Vector3f.YP.rotation(yaw))
			pose.mulPose(Vector3f.XP.rotation(pitch))

			itemInHandRenderer.renderItem(
				player,
				stack,
				ItemTransforms.TransformType.FIRST_PERSON_RIGHT_HAND,
				false,
				pose,
				event.multiBufferSource,
				event.packedLight
			)

			pose.popPose()

			item.dataTable = null
			inRender = false
		}

		private val localPlayer: Player? get() {
			return Minecraft.getInstance().player
		}

		private var firedPrimary = false
		private var firedSecondary = false

		private fun processClientInputs(item: AbstractWeaponItem<*>, itemStack: ItemStack) {
			val minecraft = Minecraft.getInstance()
			val player = minecraft.player ?: return
			val dt = item.dataTable(itemStack)

			if (!dt.wantsToScope && minecraft.options.keyUse.isDown) {
				dt.wantsToScope = true
				MatteryNetworking.CHANNEL.sendToServer(WeaponScopePacket.SCOPE_IN)
			} else if (dt.wantsToScope && !minecraft.options.keyUse.isDown) {
				dt.wantsToScope = false
				MatteryNetworking.CHANNEL.sendToServer(WeaponScopePacket.SCOPE_OUT)
			}

			if (minecraft.options.keyAttack.isDown && (item.primaryAutomatic || !firedPrimary)) {
				firedPrimary = true

				if (item.tryPrimaryFire(itemStack, player)) {
					MatteryNetworking.CHANNEL.sendToServer(WeaponFireInputPacket.PRIMARY)
				}
			} else if (!minecraft.options.keyAttack.isDown) {
				firedPrimary = false
			}

			if (minecraft.options.keyAttack.isDown && (item.secondaryAutomatic || !firedSecondary)) {
				firedSecondary = true

				if (item.trySecondaryFire(itemStack, player)) {
					MatteryNetworking.CHANNEL.sendToServer(WeaponFireInputPacket.SECONDARY)
				}
			} else if (!minecraft.options.keyAttack.isDown) {
				firedSecondary = false
			}
		}

		@SubscribeEvent
		@Suppress("unused", "unchecked_cast")
		fun tick(event: TickEvent.PlayerTickEvent) {
			if (event.phase != TickEvent.Phase.START)
				return

			if (event.side == LogicalSide.CLIENT && event.player != localPlayer)
				return

			val held = event.player.mainHandItem

			for (stack in event.player.inventory.items) {
				if (held !== stack) {
					(stack.item as? AbstractWeaponItem<*>)?.thinkHolstered(stack, event.player)
				}

				(stack.item as? AbstractWeaponItem<*>)?.let {
					if (event.side == LogicalSide.CLIENT) {
						(it as AbstractWeaponItem<WeaponDataTable>).dataTable = predictedData
					}

					it.think2(stack, event.player)

					if (event.side == LogicalSide.CLIENT) {
						it.dataTable = null
					}
				}
			}

			val prev = items[event.player]
			val heldItem = held.item

			if (held.weaponDataTable?.uuid != prev?.weaponDataTable?.uuid) {
				(prev?.item as? AbstractWeaponItem<*>)?.let {
					if (event.side == LogicalSide.CLIENT) {
						(it as AbstractWeaponItem<WeaponDataTable>).dataTable = predictedData
					}

					it.holster(prev, event.player)

					if (event.side == LogicalSide.CLIENT) {
						it.dataTable = null
					}
				}

				(held.item as? AbstractWeaponItem<*>)?.deploy(held, event.player)

				items[event.player] = held

				if (event.side == LogicalSide.CLIENT && heldItem is AbstractWeaponItem<*>) {
					predictedData = heldItem.makeDataTable(held.tagNotNull.copy())

					firedPrimary = false
					firedSecondary = false
				}
			}

			if (heldItem is AbstractWeaponItem<*>) {
				// this can fail horribly, and die horribly, if weapon item changed,
				// but its UUID did not (compound tag fraud)
				if (event.side == LogicalSide.CLIENT) {
					(heldItem as AbstractWeaponItem<WeaponDataTable>).dataTable = predictedData
					processClientInputs(heldItem, held)
				}

				heldItem.think(held, event.player)

				if (event.side == LogicalSide.CLIENT) {
					heldItem.dataTable = null
				}
			}
		}
	}
}