package ru.dbotthepony.mc.otm.item.weapon

import net.minecraft.ChatFormatting
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.Level
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.CapabilityEnergy
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.ifPresentK
import ru.dbotthepony.mc.otm.menu.FormattingHelper
import ru.dbotthepony.mc.otm.set
import kotlin.reflect.KClass

class PlasmaWeaponEnergy(val itemStack: ItemStack, private val innerCapacity: ImpreciseFraction) :
	IMatteryEnergyStorage, ICapabilityProvider, INBTSerializable<CompoundTag> {
	private val energyResolver = LazyOptional.of { this }
	private var innerBatteryLevel = ImpreciseFraction.ZERO

	var battery: ItemStack = ItemStack.EMPTY

	override fun <T : Any> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (cap === MatteryCapability.ENERGY || cap === CapabilityEnergy.ENERGY) {
			return energyResolver.cast()
		}

		return LazyOptional.empty()
	}

	override fun serializeNBT(): CompoundTag {
		val nbt = CompoundTag()
		nbt["battery_level"] = innerBatteryLevel.serializeNBT()
		nbt["battery"] = battery.serializeNBT()
		return nbt
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		innerBatteryLevel = ImpreciseFraction.deserializeNBT(nbt["battery_level"])
		battery = ItemStack.of(nbt["battery"] as CompoundTag)
	}

	override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return ImpreciseFraction.ZERO
	}

	override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		var totalExtracted = ImpreciseFraction.ZERO

		if (!battery.isEmpty) {
			battery.getCapability(CapabilityEnergy.ENERGY).ifPresentK {
				val extracted = it.extractEnergy(howMuch, simulate)

				if (extracted >= howMuch) {
					return extracted
				}

				howMuch -= extracted
				totalExtracted += extracted
			}
		}

		val newEnergy = (innerBatteryLevel - howMuch).moreThanZero()
		val diff = innerBatteryLevel - newEnergy

		if (!simulate) {
			innerBatteryLevel = newEnergy
		}

		return diff + totalExtracted
	}

	override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return receiveEnergyInner(howMuch, simulate)
	}

	override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		var totalReceived = ImpreciseFraction.ZERO

		if (!battery.isEmpty) {
			battery.getCapability(CapabilityEnergy.ENERGY).ifPresentK {
				val received = it.receiveEnergy(howMuch, simulate)

				if (received >= howMuch) {
					return received
				}

				howMuch -= received
				totalReceived += received
			}
		}

		if (innerBatteryLevel >= innerCapacity) {
			return totalReceived
		}

		val newEnergy = (innerBatteryLevel - howMuch).coerceAtMost(innerCapacity)
		val diff = newEnergy - innerBatteryLevel

		if (!simulate) {
			innerBatteryLevel = newEnergy
		}

		return diff + totalReceived
	}

	fun think() {
		if (!battery.isEmpty && innerBatteryLevel < innerCapacity) {
			battery.getCapability(CapabilityEnergy.ENERGY).ifPresentK {
				innerBatteryLevel += it.extractEnergy(innerCapacity - innerBatteryLevel, false)
			}
		}
	}

	override fun canExtract(): Boolean {
		return false
	}

	override fun canReceive(): Boolean {
		return true
	}

	override val batteryLevel: ImpreciseFraction
		get() = (battery.energy?.energyStoredMattery ?: ImpreciseFraction.ZERO) + innerBatteryLevel

	override val maxBatteryLevel: ImpreciseFraction
		get() = (battery.energy?.maxEnergyStoredMattery ?: ImpreciseFraction.ZERO) + innerCapacity
}

abstract class PlasmaWeaponItem<D : WeaponDataTable>(tables: KClass<D>, private val energyCapacity: ImpreciseFraction) : AbstractWeaponItem<D>(tables) {
	override fun appendHoverText(
		itemStack: ItemStack,
		p_41422_: Level?,
		p_41423_: MutableList<Component>,
		p_41424_: TooltipFlag
	) {
		super.appendHoverText(itemStack, p_41422_, p_41423_, p_41424_)

		itemStack.getCapability(MatteryCapability.ENERGY).ifPresentK {
			p_41423_.add(
				TranslatableComponent(
					"otm.item.power.normal.storage",
					FormattingHelper.formatPower(it.batteryLevel),
					FormattingHelper.formatPower(it.maxBatteryLevel)
				).withStyle(ChatFormatting.GRAY)
			)
		}
	}

	fun energyData(itemStack: ItemStack) = itemStack.matteryEnergy as PlasmaWeaponEnergy

	override fun think2(itemStack: ItemStack, player: Player, dt: D) {
		super.think2(itemStack, player, dt)

		itemStack.getCapability(MatteryCapability.ENERGY).ifPresentK {
			it as PlasmaWeaponEnergy
			it.think()
		}
	}

	override fun initCapabilities(stack: ItemStack, nbt: CompoundTag?): ICapabilityProvider {
		return PlasmaWeaponEnergy(stack, energyCapacity)
	}
}