package ru.dbotthepony.mc.otm.registry

import net.minecraft.core.BlockPos
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.damagesource.DamageSource
import net.minecraft.world.entity.EntityType
import net.minecraft.world.item.DyeColor
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.block.*
import net.minecraft.world.level.block.state.BlockBehaviour
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.material.Material
import net.minecraft.world.level.material.MaterialColor
import net.minecraftforge.fml.javafmlmod.FMLJavaModLoadingContext
import net.minecraftforge.registries.ForgeRegistry
import net.minecraftforge.registries.IForgeRegistry
import net.minecraftforge.registries.NewRegistryEvent
import net.minecraftforge.registries.RegistryBuilder
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import java.util.function.Supplier
import kotlin.properties.ReadOnlyProperty
import kotlin.reflect.KProperty

class CrateProperties(val color: MaterialColor, val name: String) {
	fun makeBlock(): Block {
		return Block(BlockBehaviour.Properties.of(Material.METAL, color)
			.sound(SoundType.METAL)
			.requiresCorrectToolForDrops()
			.strength(5.0F, 6.0F))
	}
}

class IndustrialGlassProperties(val color: DyeColor?, val name: String, val namePane: String) {
	constructor(color: DyeColor?) : this(color, "industrial_glass${color?.let { "_$it" } ?: ""}", "industrial_glass_pane${color?.let { "_$it" } ?: ""}")

	fun makeBlock(): Block {
		val properties = BlockBehaviour.Properties.of(Material.GLASS, if (color != null) color.materialColor else MaterialColor.NONE)
				.strength(1.5f, 5.0f)
				.requiresCorrectToolForDrops()
				.sound(SoundType.GLASS)
				.noOcclusion()
				.isValidSpawn { _: BlockState, _: BlockGetter, _: BlockPos, _: EntityType<*>? -> false }
				.isRedstoneConductor { _: BlockState, _: BlockGetter, _: BlockPos -> false }
				.isSuffocating { _: BlockState, _: BlockGetter, _: BlockPos -> false }
				.isViewBlocking { _: BlockState, _: BlockGetter, _: BlockPos -> false }

		if (color != null) {
			return StainedGlassBlock(color, properties)
		}

		return GlassBlock(properties)
	}

	fun makePaneBlock(): Block {
		val properties = BlockBehaviour.Properties.of(Material.GLASS, if (color != null) color.materialColor else MaterialColor.NONE)
			.strength(1.25f, 5.0f)
			.requiresCorrectToolForDrops()
			.sound(SoundType.GLASS)
			.noOcclusion()

		if (color != null) {
			return StainedGlassPaneBlock(color, properties)
		}

		return IronBarsBlock(properties)
	}
}

private class RegistryDelegate<V>(private val name: String) {
	private var value: Supplier<V>? = null

	fun get(): V {
		val supp = value ?: throw IllegalStateException("Tried to access uninitialized registry $name")
		return supp.get() ?: throw IllegalStateException("Accessing registry $name too early")
	}

	fun write(value: Supplier<V>) {
		if (this.value != null) {
			throw IllegalStateException("Already have registry written")
		}

		this.value = value
	}
}

object MRegistry {
	private val features = RegistryDelegate<IForgeRegistry<AndroidFeatureType<*>>>("ANDROID_FEATURES")
	private val research = RegistryDelegate<IForgeRegistry<AndroidResearchType<*>>>("ANDROID_RESEARCH")

	val ANDROID_FEATURES get() = features.get() as ForgeRegistry<AndroidFeatureType<*>>
	val ANDROID_RESEARCH get() = research.get() as ForgeRegistry<AndroidResearchType<*>>

	fun initialize(context: FMLJavaModLoadingContext) {
		context.modEventBus.addListener(this::register)
		context.modEventBus.register(MStats)
		context.modEventBus.register(LootModifiers)

		MBlocks.register()
		MBlockEntities.register()
		MEntityTypes.register()
		MMenus.register()
		MItems.register()
		AndroidFeatures.register()
		AndroidResearch.register()
	}

	private fun register(event: NewRegistryEvent) {
		features.write(RegistryBuilder<AndroidFeatureType<*>>().let {
			it.setName(ResourceLocation(OverdriveThatMatters.MOD_ID, "android_features"))
			it.type = AndroidFeatureType::class.java
			event.create(it)
		})

		research.write(RegistryBuilder<AndroidResearchType<*>>().let {
			it.setName(ResourceLocation(OverdriveThatMatters.MOD_ID, "android_research"))
			it.type = AndroidResearchType::class.java
			event.create(it)
		})
	}

	val DAMAGE_BECOME_ANDROID = ImmutableDamageSource(DamageSource("otm_become_android").bypassArmor().bypassInvul().bypassMagic())
	val DAMAGE_BECOME_HUMANE = ImmutableDamageSource(DamageSource("otm_become_humane").bypassArmor().bypassInvul().bypassMagic())
	val DAMAGE_EVENT_HORIZON = ImmutableDamageSource(DamageSource("otm_event_horizon").bypassMagic().bypassArmor())
	val DAMAGE_HAWKING_RADIATION = ImmutableDamageSource(DamageSource("otm_hawking_radiation"))
	const val DAMAGE_EMP_NAME = "otm_emp"
	const val DAMAGE_PLASMA_NAME = "otm_plasma"
	val DAMAGE_EMP: DamageSource = ImmutableDamageSource(EMPDamageSource())

	init {
		// DAMAGE_HAWKING_RADIATION.bypassMagic().bypassArmor();
	}

	val CRATE_RED = CrateProperties(MaterialColor.COLOR_RED, "crate_red")
	val CRATE_BLUE = CrateProperties(MaterialColor.COLOR_BLUE, "crate_blue")
	val CRATE_YELLOW = CrateProperties(MaterialColor.COLOR_YELLOW, "crate_yellow")
	val CRATE_GREEN = CrateProperties(MaterialColor.COLOR_GREEN, "crate_green")
	val CRATE_BLACK = CrateProperties(MaterialColor.COLOR_BLACK, "crate_black")
	val CRATE_PINK = CrateProperties(MaterialColor.COLOR_PINK, "crate_pink")
	val CRATE_PURPLE = CrateProperties(MaterialColor.COLOR_PURPLE, "crate_purple")

	val CRATE_LIST = listOf(
		CRATE_RED,
		CRATE_BLUE,
		CRATE_YELLOW,
		CRATE_GREEN,
		CRATE_BLACK,
		CRATE_PINK,
		CRATE_PURPLE,
	)

	val INDUSTRIAL_GLASS = IndustrialGlassProperties(null)
	val INDUSTRIAL_GLASS_WHITE = IndustrialGlassProperties(DyeColor.WHITE)
	val INDUSTRIAL_GLASS_ORANGE = IndustrialGlassProperties(DyeColor.ORANGE)
	val INDUSTRIAL_GLASS_MAGENTA = IndustrialGlassProperties(DyeColor.MAGENTA)
	val INDUSTRIAL_GLASS_LIGHT_BLUE = IndustrialGlassProperties(DyeColor.LIGHT_BLUE)
	val INDUSTRIAL_GLASS_YELLOW = IndustrialGlassProperties(DyeColor.YELLOW)
	val INDUSTRIAL_GLASS_LIME = IndustrialGlassProperties(DyeColor.LIME)
	val INDUSTRIAL_GLASS_PINK = IndustrialGlassProperties(DyeColor.PINK)
	val INDUSTRIAL_GLASS_GRAY = IndustrialGlassProperties(DyeColor.GRAY)
	val INDUSTRIAL_GLASS_LIGHT_GRAY = IndustrialGlassProperties(DyeColor.LIGHT_GRAY)
	val INDUSTRIAL_GLASS_CYAN = IndustrialGlassProperties(DyeColor.CYAN)
	val INDUSTRIAL_GLASS_PURPLE = IndustrialGlassProperties(DyeColor.PURPLE)
	val INDUSTRIAL_GLASS_BLUE = IndustrialGlassProperties(DyeColor.BLUE)
	val INDUSTRIAL_GLASS_BROWN = IndustrialGlassProperties(DyeColor.BROWN)
	val INDUSTRIAL_GLASS_GREEN = IndustrialGlassProperties(DyeColor.GREEN)
	val INDUSTRIAL_GLASS_RED = IndustrialGlassProperties(DyeColor.RED)
	val INDUSTRIAL_GLASS_BLACK = IndustrialGlassProperties(DyeColor.BLACK)

	val INDUSTRIAL_GLASS_LIST = listOf(
		INDUSTRIAL_GLASS,
		INDUSTRIAL_GLASS_WHITE,
		INDUSTRIAL_GLASS_ORANGE,
		INDUSTRIAL_GLASS_MAGENTA,
		INDUSTRIAL_GLASS_LIGHT_BLUE,
		INDUSTRIAL_GLASS_YELLOW,
		INDUSTRIAL_GLASS_LIME,
		INDUSTRIAL_GLASS_PINK,
		INDUSTRIAL_GLASS_GRAY,
		INDUSTRIAL_GLASS_LIGHT_GRAY,
		INDUSTRIAL_GLASS_CYAN,
		INDUSTRIAL_GLASS_PURPLE,
		INDUSTRIAL_GLASS_BLUE,
		INDUSTRIAL_GLASS_BROWN,
		INDUSTRIAL_GLASS_GREEN,
		INDUSTRIAL_GLASS_RED,
		INDUSTRIAL_GLASS_BLACK,
	)
}
