package ru.dbotthepony.mc.otm.storage

import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import java.util.*
import net.minecraftforge.registries.ForgeRegistry

interface IStorageStack {
	fun copy(): IStorageStack

	/**
	 * Despite count being [ImpreciseFraction], far not all items can be "fraction"
	 * into pieces (e.g. ItemStacks).
	 *
	 * Consult [StorageStackType.fractional] to see if item has actual fractions.
	 *
	 * Implementation MUST [ImpreciseFraction.floor] received value if fractions
	 * are not supported, and NOT throw any errors. Writing fractional value is a correct
	 * behavior, and is used in base mod.
	 *
	 * Behavior of writing negative value is undefined.
	 */
	var count: ImpreciseFraction
	val isEmpty: Boolean

	/**
	 * @return max stack size for this stack object,
	 * null if unlimited (default)
	 */
	val maxStackSize: ImpreciseFraction? get() = null

	/**
	 * In base mod, a [hashCode] but for use with AVL trees.
	 *
	 * This property represent (almost) unique key used to quickly look up for similar stacks.
	 * This property **must** be equal for two entries if [sameItem] returns true.
	 * This value should not change.
	 *
	 * A good example of key is [ForgeRegistry.getID].
	 *
	 * Users of this property should prefer binary search trees (e.g. AVL) with this value as key.
	 */
	val longHashCode: Long

	/**
	 * [other] is the stack to compare.
	 *
	 * If not the same of this object type just return false
	 * Returns boolean representing whenever internal state of *this* equals to [other];
	 *
	 * This includes (nbt) tags, mapping IDs, capabilities, **excluding** amount;
	 *
	 * Example of this behavior is demonstrated by ItemStack.isSameItemSameTags.
	 */
	fun sameItem(other: IStorageStack): Boolean

	/**
	 * Increase [count] by [amount]
	 */
	fun grow(amount: ImpreciseFraction) {
		count += amount
	}

	/**
	 * Decrease [count] by [amount]
	 */
	fun shrink(amount: ImpreciseFraction) {
		count -= amount
	}
}

/**
 * Storage system root, along IStorageStack interface
 */
interface IStorage<T : IStorageStack> {
	/**
	 * @return Identity of this virtual component
	 */
	val storageType: StorageStackType<T>
}

interface IStorageTrigger<T : IStorageStack> : IStorage<T> {
	/**
	 * [listener] is [IStorageListener] which want to subscribe to our events
	 */
	fun addListener(listener: IStorageListener<T>): Boolean

	/**
	 * [listener] is [IStorageListener] which want to unsubscribe from our events
	 */
	fun removeListener(listener: IStorageListener<T>): Boolean
}

interface IStorageConsumer<T : IStorageStack> : IStorage<T> {
	/**
	 * Inserts an item into system.
	 * @return leftover, might equal to [stack] if no items were inserted
	 */
	fun insertStack(stack: T, simulate: Boolean): T
}

interface IStorageView<T : IStorageStack> : IStorageTrigger<T> {
	/**
	 * @param id identifier of stack
	 * @return stored object (not a copy). Do not edit it.
	 */
	fun getStack(id: UUID): T

	/**
	 * If tuple does not exist, returns empty stack
	 *
	 * @param id identifier of stack to extract
	 * @param amount amount of units to extract
	 * @param simulate whenever to simulate the action or not
	 * @return copy of object, with amount of units actually extracted
	 */
	fun extractStack(id: UUID, amount: ImpreciseFraction, simulate: Boolean): T
	fun extractStack(id: UUID, amount: Int, simulate: Boolean): T = extractStack(id, ImpreciseFraction(amount), simulate)
	fun extractStack(id: UUID, amount: Long, simulate: Boolean): T = extractStack(id, ImpreciseFraction(amount), simulate)

	/**
	 * Designed for views, for extraction with less computation overhead caused by
	 * copying stack extracted
	 *
	 * @param id identifier of stack to extract
	 * @param amount desired amount to extract
	 * @param simulate whenever to simulate the action or not
	 * @return amount extracted
	 */
	fun extractStackCount(id: UUID, amount: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return extractStack(id, amount, simulate).count
	}

	fun extractStackCount(id: UUID, amount: Int, simulate: Boolean): ImpreciseFraction = extractStackCount(id, ImpreciseFraction(amount), simulate)
	fun extractStackCount(id: UUID, amount: Long, simulate: Boolean): ImpreciseFraction = extractStackCount(id, ImpreciseFraction(amount), simulate)

	fun getStacks(): Collection<IStorageTuple<T>>

	fun addListenerAuto(listener: IStorageListener<T>): Boolean {
		if (addListener(listener)) {
			for (stack in getStacks()) {
				listener.addStack(stack.stack, stack.id, this)
			}

			return true
		}

		return false
	}

	fun removeListenerAuto(listener: IStorageListener<T>): Boolean {
		if (removeListener(listener)) {
			for (stack in getStacks()) {
				listener.removeStack(stack.stack, stack.id)
			}

			return true
		}

		return false
	}
}

interface IStorageListener<T : IStorageStack> {
	/**
	 * Fired on whenever an object is added (to listener) we subscribed to
	 */
	fun addStack(stack: T, id: UUID, provider: IStorageView<T>)

	fun addStack(tuple: IStorageTuple<T>, provider: IStorageView<T>) {
		addStack(tuple.stack, tuple.id, provider)
	}

	/**
	 * Fired on whenever an object is changes on listener we subscribed to
	 */
	fun changeStack(stack: T, id: UUID, oldCount: ImpreciseFraction)

	fun changeStack(tuple: IStorageTuple<T>, oldCount: ImpreciseFraction) {
		changeStack(tuple.stack, tuple.id, oldCount)
	}

	/**
	 * Fired on whenever an object is removed from listener we subscribed to
	 */
	fun removeStack(stack: T, id: UUID)

	fun removeStack(tuple: IStorageTuple<T>) {
		removeStack(tuple.stack, tuple.id)
	}
}

interface IStorageTuple<T : IStorageStack> {
	val id: UUID
	val stack: T
}

class StorageTuple<T : IStorageStack>(override val id: UUID, override val stack: T) : IStorageTuple<T>
interface IStorageComponent<T : IStorageStack> : IStorageView<T>, IStorageConsumer<T>

/**
 * Component which (most time) proxy other components (combine their contents into single view)
 */
interface IVirtualStorageComponent<T : IStorageStack> : IStorageComponent<T>, IStorageListener<T> {
	fun add(identity: IStorage<T>)
	fun remove(identity: IStorage<T>)
	fun contains(identity: IStorage<T>): Boolean
}
