package ru.dbotthepony.mc.otm.storage

import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.Fraction
import java.util.IdentityHashMap

open class StorageStackType<T : IStorageStack>(
	val identity: Class<T>,
	open val empty: T,

	/**
	 * Speculated energy required per operation on stack with size of 1
	 */
	open val energyPerOperation: ImpreciseFraction,

	/**
	 * Whenever is this stack supports fractional part (e.g. 0.5).
	 *
	 * Keep in mind fractions are imprecise and can lead to rounding errors.
	 * [ImpreciseFraction] class attempts to negate most of the issues
	 * (e.g. 0.1 + 0.2 eventually getting its 0....4 part into whole part),
	 * but that is about it.
	 *
	 * On design side note, storage system could have been using [Fraction], but there is an issue:
	 * they are **precise**. Under precise, means that anything that continuously divide/multiply them
	 * they become more and more "irrational", greatly slowing down operations. Worst case scenario:
	 * value is getting divided by [Long.MAX_VALUE] again and again, creating insanely huge divisor.
	 */
	open val fractional: Boolean
) {
	open fun energyPerOperation(stack: T): ImpreciseFraction {
		return stack.count * energyPerOperation
	}

	open fun energyPerStorage(stack: T) = energyPerOperation(stack)
	open fun energyPerExtraction(stack: T) = energyPerOperation(stack)
	open fun energyForUpkeep(stack: T) = ImpreciseFraction.ZERO
}

@Suppress("unchecked_cast")
object StorageRegistry {
	private val REGISTRY = IdentityHashMap<Class<*>, StorageStackType<*>>()

	@JvmStatic
	fun register(type: StorageStackType<*>): StorageStackType<*> {
		check(!REGISTRY.containsKey(type.identity)) { "Already have storage stack type for ${type.identity}" }
		REGISTRY[type.identity] = type
		return type
	}

	/**
	 * Refer to [StorageStackType] for explanation of flags.
	 */
	@JvmStatic
	fun <T : IStorageStack> register(
		identity: Class<T>,
		empty: T,
		energyPerOperation: ImpreciseFraction,
		fractional: Boolean
	): StorageStackType<T> {
		return register(StorageStackType(identity, empty, energyPerOperation, fractional)) as StorageStackType<T>
	}

	@JvmStatic
	fun <T : IStorageStack> get(identity: Class<T>): StorageStackType<T> {
		val get = REGISTRY[identity] ?: throw NoSuchElementException("Registry does not contain $identity")
		return get as StorageStackType<T>
	}

	@JvmStatic
	fun <T : IStorageStack> getOrNull(identity: Class<T>): StorageStackType<T>? {
		return REGISTRY[identity] as StorageStackType<T>?
	}
}
